"""
Tests for MB04VX: Separate epsilon and infinite pencils from staircase form.

MB04VX separates s*E(eps)-A(eps) and s*E(inf)-A(inf) from s*E(eps,inf)-A(eps,inf).
This is similar to MB04TX but with NBLCKS as input-only and MNEI having 3 elements.
"""

import numpy as np
import pytest
from slicot import mb04vx


class TestMB04VXBasic:
    """Basic functionality tests for MB04VX."""

    def test_simple_2x3_pencil(self):
        """
        Test with simple 2x3 pencil with 2 blocks.

        Random seed: 42 (for reproducibility)
        """
        np.random.seed(42)

        m, n = 3, 4
        nblcks = 2

        # Create staircase form pencil (already processed by MB04UD/MB04TX-like)
        # Block structure: nu = [2, 1], mu = [2, 2]
        inuk = np.array([2, 1], dtype=np.int32)
        imuk = np.array([2, 2], dtype=np.int32)

        # A matrix in staircase form
        a = np.array([
            [1.0, 2.0, 0.5, 0.3],
            [0.0, 3.0, 1.0, 0.2],
            [0.0, 0.0, 2.0, 0.1]
        ], order='F', dtype=float)

        # E matrix in staircase form (upper triangular structure in blocks)
        e = np.array([
            [1.0, 0.5, 0.3, 0.2],
            [0.0, 1.0, 0.4, 0.1],
            [0.0, 0.0, 1.0, 0.5]
        ], order='F', dtype=float)

        # Transformation matrices
        q = np.eye(m, order='F', dtype=float)
        z = np.eye(n, order='F', dtype=float)

        # Call routine
        a_out, e_out, q_out, z_out, inuk_out, imuk_out, mnei = mb04vx(
            a, e, q, z, nblcks, inuk, imuk, updatq=True, updatz=True
        )

        # Validate MNEI output dimensions
        assert len(mnei) == 3
        meps, neps, minf = mnei[0], mnei[1], mnei[2]

        # Dimensions should be consistent
        assert meps >= 0
        assert neps >= 0
        assert minf >= 0

        # After separation, the sum of inuk should equal meps
        # and sum of imuk should equal neps
        assert np.sum(inuk_out) == meps
        assert np.sum(imuk_out) == neps

    def test_identity_transformations(self):
        """
        Test that Q and Z are orthogonal when updated.

        Random seed: 123 (for reproducibility)
        """
        np.random.seed(123)

        m, n = 4, 5
        nblcks = 2

        inuk = np.array([2, 2], dtype=np.int32)
        imuk = np.array([3, 2], dtype=np.int32)

        # Create matrices
        a = np.random.randn(m, n).astype(float, order='F')
        e = np.random.randn(m, n).astype(float, order='F')
        q = np.eye(m, order='F', dtype=float)
        z = np.eye(n, order='F', dtype=float)

        a_out, e_out, q_out, z_out, inuk_out, imuk_out, mnei = mb04vx(
            a, e, q, z, nblcks, inuk, imuk, updatq=True, updatz=True
        )

        # Q should remain orthogonal
        qtq = q_out.T @ q_out
        np.testing.assert_allclose(qtq, np.eye(m), rtol=1e-14, atol=1e-14)

        # Z should remain orthogonal
        ztz = z_out.T @ z_out
        np.testing.assert_allclose(ztz, np.eye(n), rtol=1e-14, atol=1e-14)

    def test_no_update_q_z(self):
        """
        Test with UPDATQ=False and UPDATZ=False.

        Random seed: 456 (for reproducibility)
        """
        np.random.seed(456)

        m, n = 3, 4
        nblcks = 2

        inuk = np.array([2, 1], dtype=np.int32)
        imuk = np.array([2, 2], dtype=np.int32)

        a = np.random.randn(m, n).astype(float, order='F')
        e = np.random.randn(m, n).astype(float, order='F')

        # Dummy Q and Z (1x1)
        q = np.array([[1.0]], order='F', dtype=float)
        z = np.array([[1.0]], order='F', dtype=float)

        # Call without updating Q/Z
        a_out, e_out, q_out, z_out, inuk_out, imuk_out, mnei = mb04vx(
            a, e, q, z, nblcks, inuk, imuk, updatq=False, updatz=False
        )

        # Q and Z should be None when not updated
        assert q_out is None
        assert z_out is None

        # MNEI should still be valid
        assert len(mnei) == 3


class TestMB04VXEdgeCases:
    """Edge case tests for MB04VX."""

    def test_zero_dimensions(self):
        """Test with M=0 or N=0."""
        m, n = 0, 0
        nblcks = 0

        inuk = np.array([], dtype=np.int32)
        imuk = np.array([], dtype=np.int32)

        a = np.array([[]], order='F', dtype=float).reshape(0, 0)
        e = np.array([[]], order='F', dtype=float).reshape(0, 0)
        q = np.array([[1.0]], order='F', dtype=float)
        z = np.array([[1.0]], order='F', dtype=float)

        a_out, e_out, q_out, z_out, inuk_out, imuk_out, mnei = mb04vx(
            a, e, q, z, nblcks, inuk, imuk, updatq=False, updatz=False
        )

        # MNEI should be all zeros for empty input
        assert mnei[0] == 0
        assert mnei[1] == 0
        assert mnei[2] == 0

    def test_single_block(self):
        """
        Test with single block (nblcks=1).

        Random seed: 789 (for reproducibility)
        """
        np.random.seed(789)

        m, n = 2, 3
        nblcks = 1

        inuk = np.array([2], dtype=np.int32)
        imuk = np.array([3], dtype=np.int32)

        a = np.random.randn(m, n).astype(float, order='F')
        e = np.random.randn(m, n).astype(float, order='F')
        q = np.eye(m, order='F', dtype=float)
        z = np.eye(n, order='F', dtype=float)

        a_out, e_out, q_out, z_out, inuk_out, imuk_out, mnei = mb04vx(
            a, e, q, z, nblcks, inuk, imuk, updatq=True, updatz=True
        )

        # Results should be valid
        assert len(mnei) == 3
        assert mnei[0] >= 0
        assert mnei[1] >= 0
        assert mnei[2] >= 0


class TestMB04VXNumerical:
    """Numerical property tests for MB04VX."""

    def test_transformation_consistency(self):
        """
        Test that Q' * A_orig * Z = A_out relationship holds.

        Random seed: 999 (for reproducibility)
        """
        np.random.seed(999)

        m, n = 4, 5
        nblcks = 2

        inuk = np.array([2, 2], dtype=np.int32)
        imuk = np.array([3, 2], dtype=np.int32)

        a_orig = np.random.randn(m, n).astype(float, order='F')
        e_orig = np.random.randn(m, n).astype(float, order='F')
        q = np.eye(m, order='F', dtype=float)
        z = np.eye(n, order='F', dtype=float)

        a_out, e_out, q_out, z_out, inuk_out, imuk_out, mnei = mb04vx(
            a_orig.copy(order='F'), e_orig.copy(order='F'),
            q, z, nblcks, inuk.copy(), imuk.copy(),
            updatq=True, updatz=True
        )

        # Q' * A_orig * Z should equal A_out
        a_transformed = q_out.T @ a_orig @ z_out
        np.testing.assert_allclose(a_transformed, a_out, rtol=1e-13, atol=1e-14)

        # Q' * E_orig * Z should equal E_out
        e_transformed = q_out.T @ e_orig @ z_out
        np.testing.assert_allclose(e_transformed, e_out, rtol=1e-13, atol=1e-14)

    def test_determinant_preservation(self):
        """
        Test that det(Q) and det(Z) have magnitude 1 (orthogonal).

        Random seed: 888 (for reproducibility)
        """
        np.random.seed(888)

        m, n = 4, 5
        nblcks = 2

        inuk = np.array([2, 2], dtype=np.int32)
        imuk = np.array([3, 2], dtype=np.int32)

        a = np.random.randn(m, n).astype(float, order='F')
        e = np.random.randn(m, n).astype(float, order='F')
        q = np.eye(m, order='F', dtype=float)
        z = np.eye(n, order='F', dtype=float)

        a_out, e_out, q_out, z_out, inuk_out, imuk_out, mnei = mb04vx(
            a, e, q, z, nblcks, inuk, imuk, updatq=True, updatz=True
        )

        # |det(Q)| = 1 for orthogonal matrix
        det_q = np.linalg.det(q_out)
        np.testing.assert_allclose(abs(det_q), 1.0, rtol=1e-14)

        # |det(Z)| = 1 for orthogonal matrix
        det_z = np.linalg.det(z_out)
        np.testing.assert_allclose(abs(det_z), 1.0, rtol=1e-14)



"""
Tests for TB01TD: Balance state-space representation.

Reduces (A,B,C,D) to balanced form using state permutations and scalings.
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose


class TestTB01TD:
    """Tests for tb01td routine."""

    def test_html_example(self):
        """
        Test case from SLICOT HTML documentation.

        N=5, M=2, P=2 system.
        A is read row-by-row: ((A(I,J), J=1,N), I=1,N)
        B is read column-by-column: ((B(I,J), I=1,N), J=1,M)
        C is read row-by-row: ((C(I,J), J=1,N), I=1,P)
        D is read row-by-row: ((D(I,J), J=1,M), I=1,P)

        Note: DGEBAL can produce different valid balancing results.
        We validate by checking eigenvalue and transfer function preservation.
        """
        from slicot import tb01td

        n, m, p = 5, 2, 2

        # A matrix (5x5) - read row-by-row
        a = np.array([
            [  0.0,   0.0,   1.0,   4.0,   5.0],
            [ 50.0,  10.0,   1.0,   0.0,   0.0],
            [  0.0,   0.0,  90.0,  10.0,   0.0],
            [  0.0,   1.0,   1.0,   1.0,   1.0],
            [100.0,   0.0,   0.0,   0.0,  70.0]
        ], dtype=float, order='F')

        # B matrix (5x2) - read column-by-column
        b = np.array([
            [0.0,   0.0],
            [2.0,  20.0],
            [0.0, 100.0],
            [1.0,   1.0],
            [2.0,   0.0]
        ], dtype=float, order='F')

        # C matrix (2x5) - read row-by-row
        c = np.array([
            [1.0, 0.0, 0.0, 1.0, 0.0],
            [1.0, 1.0, 0.0, 2.0, 1.0]
        ], dtype=float, order='F')

        # D matrix (2x2) - read row-by-row
        d = np.array([
            [1.0, 1.0],
            [1.0, 1.0]
        ], dtype=float, order='F')

        # Keep copies for comparison
        a_orig = a.copy()
        b_orig = b.copy()
        c_orig = c.copy()
        d_orig = d.copy()
        eig_orig = np.linalg.eigvals(a_orig)

        # Call routine
        low, igh, scstat, scin, scout, info = tb01td(n, m, p, a, b, c, d)

        assert info == 0
        assert low == 1
        assert igh == 5

        # Verify array dimensions
        assert len(scstat) == n
        assert len(scin) == m
        assert len(scout) == p

        # Eigenvalues must be preserved (similarity transformation)
        eig_bal = np.linalg.eigvals(a)
        assert_allclose(sorted(eig_orig.real), sorted(eig_bal.real), rtol=1e-12)

        # Transfer function must be preserved at multiple s values
        I = np.eye(n)
        for s in [1.0, 10.0, 1.0 + 1.0j]:
            # Original: G(s) = C*inv(sI-A)*B + D
            sI_minus_A_orig = s * I - a_orig
            G_orig = c_orig @ np.linalg.solve(sI_minus_A_orig, b_orig) + d_orig

            # Balanced: Reconstruct original G from scaled matrices
            sI_minus_A_bal = s * I - a
            G_bal_raw = c @ np.linalg.solve(sI_minus_A_bal, b) + d
            G_reconstructed = np.diag(1.0 / scout) @ G_bal_raw @ np.diag(scin)

            assert_allclose(G_orig, G_reconstructed, rtol=1e-12)

    def test_identity_preservation(self):
        """
        Test that balancing preserves system I/O relationship.

        For balanced system (Ab, Bb, Cb, Db), the transfer function
        is preserved: C*inv(sI-A)*B + D should be unchanged.

        Uses impulse response equivalence at multiple s values.

        Random seed: 42 (for reproducibility)
        """
        from slicot import tb01td

        np.random.seed(42)
        n, m, p = 3, 2, 2

        # Create original system matrices
        a = np.array([
            [-1.0,  0.5, -0.2],
            [ 0.3, -2.0,  0.4],
            [ 0.1, -0.3, -0.5]
        ], dtype=float, order='F')

        b = np.array([
            [1.0, 0.5],
            [0.2, 1.0],
            [0.3, 0.1]
        ], dtype=float, order='F')

        c = np.array([
            [1.0, 0.3, 0.2],
            [0.1, 1.0, 0.5]
        ], dtype=float, order='F')

        d = np.array([
            [0.1, 0.0],
            [0.0, 0.1]
        ], dtype=float, order='F')

        # Keep copies for transfer function comparison
        a_orig = a.copy()
        b_orig = b.copy()
        c_orig = c.copy()
        d_orig = d.copy()

        # Balance the system
        low, igh, scstat, scin, scout, info = tb01td(n, m, p, a, b, c, d)
        assert info == 0

        # Compare transfer functions at several s values
        I = np.eye(n)
        for s in [1.0, 2.0j, 1.0 + 1.0j]:
            # Original: G(s) = C*inv(sI-A)*B + D
            sI_minus_A_orig = s * I - a_orig
            G_orig = c_orig @ np.linalg.solve(sI_minus_A_orig, b_orig) + d_orig

            # Balanced: Gb(s) = Cb*inv(sI-Ab)*Bb + Db
            # But outputs scaled by scout, inputs by 1/scin
            # True output = diag(scout) * balanced_output
            # True input needs division by scin
            sI_minus_A_bal = s * I - a
            G_bal_raw = c @ np.linalg.solve(sI_minus_A_bal, b) + d

            # Reconstruct: scale back
            # The routine scales such that:
            # d_bal(i,j) = d_orig(i,j) * scout(i) / scin(j)
            # So G_orig = diag(1/scout) @ G_bal @ diag(scin)
            G_reconstructed = np.diag(1.0 / scout) @ G_bal_raw @ np.diag(scin)

            assert_allclose(G_orig, G_reconstructed, rtol=1e-12)

    def test_zero_dimension(self):
        """Test quick return when N=M=P=0."""
        from slicot import tb01td

        n, m, p = 0, 0, 0
        a = np.zeros((1, 1), dtype=float, order='F')
        b = np.zeros((1, 1), dtype=float, order='F')
        c = np.zeros((1, 1), dtype=float, order='F')
        d = np.zeros((1, 1), dtype=float, order='F')

        low, igh, scstat, scin, scout, info = tb01td(n, m, p, a, b, c, d)

        assert info == 0
        assert low == 1  # Fortran 1-based
        assert igh == 0  # N=0

    def test_siso_system(self):
        """
        Test simple SISO system balancing.

        Random seed: 123 (for reproducibility)
        """
        from slicot import tb01td

        np.random.seed(123)
        n, m, p = 2, 1, 1

        a = np.array([
            [-0.5, 1.0],
            [ 0.0, -1.0]
        ], dtype=float, order='F')

        b = np.array([
            [1.0],
            [0.5]
        ], dtype=float, order='F')

        c = np.array([
            [1.0, 0.0]
        ], dtype=float, order='F')

        d = np.array([
            [0.0]
        ], dtype=float, order='F')

        low, igh, scstat, scin, scout, info = tb01td(n, m, p, a, b, c, d)

        assert info == 0
        # LOW and IGH should be valid indices
        assert 1 <= low <= n
        assert low <= igh <= n

        # scstat should be populated
        assert len(scstat) == n
        assert len(scin) == m
        assert len(scout) == p

    def test_eigenvalue_preservation(self):
        """
        Test that eigenvalues of A are preserved after balancing.

        Balancing uses similarity transformation: eigenvalues unchanged.

        Random seed: 456 (for reproducibility)
        """
        from slicot import tb01td

        np.random.seed(456)
        n, m, p = 4, 2, 2

        # Create a matrix with known eigenvalues
        a = np.array([
            [-1.0,  2.0,  0.0,  0.0],
            [ 0.0, -2.0,  3.0,  0.0],
            [ 0.0,  0.0, -3.0,  4.0],
            [ 0.0,  0.0,  0.0, -4.0]
        ], dtype=float, order='F')

        b = np.random.randn(n, m).astype(float, order='F')
        c = np.random.randn(p, n).astype(float, order='F')
        d = np.random.randn(p, m).astype(float, order='F')

        eig_orig = np.linalg.eigvals(a)

        low, igh, scstat, scin, scout, info = tb01td(n, m, p, a, b, c, d)

        assert info == 0

        eig_bal = np.linalg.eigvals(a)

        # Eigenvalues should be preserved (sort for comparison)
        assert_allclose(sorted(eig_orig.real), sorted(eig_bal.real), rtol=1e-12)
        assert_allclose(sorted(eig_orig.imag), sorted(eig_bal.imag), rtol=1e-12)

    def test_error_invalid_n(self):
        """Test error handling for invalid N < 0."""
        from slicot import tb01td

        n, m, p = -1, 1, 1
        a = np.zeros((1, 1), dtype=float, order='F')
        b = np.zeros((1, 1), dtype=float, order='F')
        c = np.zeros((1, 1), dtype=float, order='F')
        d = np.zeros((1, 1), dtype=float, order='F')

        low, igh, scstat, scin, scout, info = tb01td(n, m, p, a, b, c, d)

        assert info == -1



"""
Test MD03BF - FCN callback for Kowalik-Osborne nonlinear least squares problem.

MD03BF is the example FCN routine for solving a standard nonlinear least squares
problem using MD03BD. It implements the Kowalik-Osborne test function from MINPACK.

Problem: minimize sum_{i=1}^{15} (y_i - (x_1 + i/(x_2*(16-i) + x_3*tmp3)))^2
where tmp3 = (16-i) if i > 8, else i

The test data Y is from the MINPACK LMDER example:
Y = [0.14, 0.18, 0.22, 0.25, 0.29, 0.32, 0.35, 0.39, 0.37, 0.58,
     0.73, 0.96, 1.34, 2.10, 4.39]
"""
import numpy as np
import pytest

try:
    from slicot import md03bf
    HAS_SLICOT = True
except ImportError:
    HAS_SLICOT = False


def compute_kowalik_osborne_error(x):
    """
    Compute Kowalik-Osborne error function values.

    e(i) = y(i) - (x[0] + i/(x[1]*(16-i) + x[2]*tmp3))
    where tmp3 = (16-i) if i > 8, else i

    This is the reference implementation in Python.
    """
    y = np.array([0.14, 0.18, 0.22, 0.25, 0.29,
                  0.32, 0.35, 0.39, 0.37, 0.58,
                  0.73, 0.96, 1.34, 2.10, 4.39], dtype=np.float64)
    m = 15
    e = np.zeros(m, dtype=np.float64)
    for i in range(1, m + 1):
        tmp1 = float(i)
        tmp2 = float(16 - i)
        tmp3 = tmp2 if i > 8 else tmp1
        e[i-1] = y[i-1] - (x[0] + tmp1 / (x[1]*tmp2 + x[2]*tmp3))
    return e


def compute_kowalik_osborne_jacobian(x):
    """
    Compute Kowalik-Osborne Jacobian matrix.

    J(i,1) = -1
    J(i,2) = i*(16-i) / (x[1]*(16-i) + x[2]*tmp3)^2
    J(i,3) = i*tmp3 / (x[1]*(16-i) + x[2]*tmp3)^2

    This is the reference implementation in Python.
    """
    m = 15
    n = 3
    j = np.zeros((m, n), dtype=np.float64, order='F')
    for i in range(1, m + 1):
        tmp1 = float(i)
        tmp2 = float(16 - i)
        tmp3 = tmp2 if i > 8 else tmp1
        tmp4 = (x[1]*tmp2 + x[2]*tmp3)**2
        j[i-1, 0] = -1.0
        j[i-1, 1] = tmp1*tmp2/tmp4
        j[i-1, 2] = tmp1*tmp3/tmp4
    return j


@pytest.mark.skipif(not HAS_SLICOT, reason="SLICOT not available")
class TestMd03bf:
    """Tests for MD03BF Kowalik-Osborne FCN callback."""

    def test_error_function_at_initial_point(self):
        """
        Test error function evaluation at initial point x = [1, 1, 1].

        Verifies IFLAG=1 computes correct error function values.
        """
        x = np.array([1.0, 1.0, 1.0], dtype=np.float64)
        e, info = md03bf(1, x)

        assert info == 0
        e_expected = compute_kowalik_osborne_error(x)
        np.testing.assert_allclose(e, e_expected, rtol=1e-14)

    def test_error_function_at_solution(self):
        """
        Test error function evaluation near the optimal solution.

        Solution from MD03BD.html: x = [0.0824, 1.1330, 2.3437]
        Final residual norm: 0.9063596E-01
        """
        x = np.array([0.0824, 1.1330, 2.3437], dtype=np.float64)
        e, info = md03bf(1, x)

        assert info == 0
        e_expected = compute_kowalik_osborne_error(x)
        np.testing.assert_allclose(e, e_expected, rtol=1e-14)

        residual_norm = np.linalg.norm(e)
        np.testing.assert_allclose(residual_norm, 0.09063596, rtol=1e-3)

    def test_jacobian_at_initial_point(self):
        """
        Test Jacobian evaluation at initial point x = [1, 1, 1].

        Verifies IFLAG=2 computes correct Jacobian matrix.
        """
        x = np.array([1.0, 1.0, 1.0], dtype=np.float64)
        j, e, nfevl, info = md03bf(2, x)

        assert info == 0
        assert nfevl == 0

        j_expected = compute_kowalik_osborne_jacobian(x)
        np.testing.assert_allclose(j, j_expected, rtol=1e-14)

    def test_jacobian_at_random_point(self):
        """
        Test Jacobian against finite differences at random point.

        Random seed: 42 (for reproducibility)
        """
        np.random.seed(42)
        x = np.random.uniform(0.5, 2.0, 3).astype(np.float64)

        j, e, nfevl, info = md03bf(2, x)

        assert info == 0

        j_expected = compute_kowalik_osborne_jacobian(x)
        np.testing.assert_allclose(j, j_expected, rtol=1e-14)

        eps = 1e-7
        m, n = 15, 3
        j_fd = np.zeros((m, n), dtype=np.float64, order='F')
        e0 = compute_kowalik_osborne_error(x)
        for k in range(n):
            x_pert = x.copy()
            x_pert[k] += eps
            e_pert = compute_kowalik_osborne_error(x_pert)
            j_fd[:, k] = (e_pert - e0) / eps

        np.testing.assert_allclose(j, j_fd, rtol=1e-5, atol=1e-6)

    def test_workspace_query(self):
        """
        Test IFLAG=3 returns correct workspace requirements.

        Expected: ipar = (M*N, 0, 0, 4*N+1, 4*N)
        """
        m, n = 15, 3
        x = np.zeros(n, dtype=np.float64)
        ipar, info = md03bf(3, x)

        assert info == 0
        assert ipar[0] == m * n
        assert ipar[1] == 0
        assert ipar[2] == 0
        assert ipar[3] == 4*n + 1
        assert ipar[4] == 4*n

    def test_jacobian_first_column_all_minus_one(self):
        """
        Mathematical property: first column of Jacobian is always -1.

        J(:,1) = de/dx1 = -1 for all i
        Random seed: 123 (for reproducibility)
        """
        np.random.seed(123)
        m = 15

        for _ in range(5):
            x = np.random.uniform(0.1, 5.0, 3).astype(np.float64)
            j, e, nfevl, info = md03bf(2, x)

            assert info == 0
            np.testing.assert_allclose(j[:, 0], np.full(m, -1.0), rtol=1e-15)



"""
Tests for MB01RX: Triangular symmetric rank-k update.

Computes R = alpha*R + beta*op(A)*B or R = alpha*R + beta*B*op(A)
where only upper or lower triangle is computed/stored.
"""

import numpy as np
import pytest


def test_mb01rx_side_left_uplo_upper_trans_no():
    """
    Test SIDE='L', UPLO='U', TRANS='N': R = alpha*R + beta*A*B (upper triangle).

    Random seed: 42 (for reproducibility)
    """
    from slicot import mb01rx

    np.random.seed(42)
    m, n = 3, 2
    alpha, beta = 2.0, 0.5

    # Input matrices (column-major)
    r = np.array([
        [4.0, 1.0, 2.0],
        [0.0, 3.0, 1.0],
        [0.0, 0.0, 5.0]
    ], order='F', dtype=float)

    a = np.array([
        [1.0, 2.0],
        [3.0, 1.0],
        [2.0, 4.0]
    ], order='F', dtype=float)

    b = np.array([
        [2.0, 1.0, 3.0],
        [1.0, 2.0, 1.0]
    ], order='F', dtype=float)

    # Expected: R_new[i,j] = alpha*R[i,j] + beta*sum_k(A[i,k]*B[k,j]) for i <= j
    # Only upper triangle computed
    r_expected = np.array([
        [10.0, 4.5, 6.5],
        [0.0, 8.5, 7.0],
        [0.0, 0.0, 15.0]
    ], order='F', dtype=float)

    r_out, info = mb01rx('L', 'U', 'N', m, n, alpha, beta, r, a, b)

    assert info == 0
    # Upper triangle only
    np.testing.assert_allclose(np.triu(r_out), np.triu(r_expected), rtol=1e-14)


def test_mb01rx_side_left_uplo_lower_trans_yes():
    """
    Test SIDE='L', UPLO='L', TRANS='T': R = alpha*R + beta*A'*B (lower triangle).

    Random seed: 123 (for reproducibility)
    """
    from slicot import mb01rx

    np.random.seed(123)
    m, n = 3, 4
    alpha, beta = 1.5, -1.0

    r = np.array([
        [2.0, 0.0, 0.0],
        [1.0, 3.0, 0.0],
        [2.0, 1.0, 4.0]
    ], order='F', dtype=float)

    a = np.array([
        [1.0, 2.0, 1.0],
        [2.0, 1.0, 3.0],
        [1.0, 2.0, 2.0],
        [3.0, 1.0, 2.0]
    ], order='F', dtype=float)

    b = np.array([
        [1.0, 2.0, 1.0],
        [2.0, 1.0, 3.0],
        [1.0, 2.0, 2.0],
        [3.0, 1.0, 2.0]
    ], order='F', dtype=float)

    # Expected: R_new[i,j] = alpha*R[i,j] + beta*sum_k(A[k,i]*B[k,j]) for i >= j
    r_expected = np.array([
        [-12.0, 0.0, 0.0],
        [-7.5, -5.5, 0.0],
        [-12.0, -9.5, -12.0]
    ], order='F', dtype=float)

    r_out, info = mb01rx('L', 'L', 'T', m, n, alpha, beta, r, a, b)

    assert info == 0
    # Lower triangle only
    np.testing.assert_allclose(np.tril(r_out), np.tril(r_expected), rtol=1e-14)


def test_mb01rx_side_right_uplo_upper_trans_no():
    """
    Test SIDE='R', UPLO='U', TRANS='N': R = alpha*R + beta*B*A (upper triangle).

    Random seed: 456 (for reproducibility)
    """
    from slicot import mb01rx

    np.random.seed(456)
    m, n = 3, 2
    alpha, beta = 0.5, 2.0

    r = np.array([
        [1.0, 2.0, 1.0],
        [0.0, 2.0, 3.0],
        [0.0, 0.0, 1.0]
    ], order='F', dtype=float)

    b = np.array([
        [1.0, 2.0],
        [2.0, 1.0],
        [1.0, 3.0]
    ], order='F', dtype=float)

    a = np.array([
        [2.0, 1.0, 3.0],
        [1.0, 2.0, 1.0]
    ], order='F', dtype=float)

    # Expected: R_new[i,j] = alpha*R[i,j] + beta*sum_k(B[i,k]*A[k,j]) for i <= j
    r_expected = np.array([
        [8.5, 11.0, 10.5],
        [0.0, 9.0, 15.5],
        [0.0, 0.0, 12.5]
    ], order='F', dtype=float)

    r_out, info = mb01rx('R', 'U', 'N', m, n, alpha, beta, r, a, b)

    assert info == 0
    np.testing.assert_allclose(np.triu(r_out), np.triu(r_expected), rtol=1e-14)


def test_mb01rx_side_right_uplo_lower_trans_yes():
    """
    Test SIDE='R', UPLO='L', TRANS='T': R = alpha*R + beta*B*A' (lower triangle).

    Random seed: 789 (for reproducibility)
    """
    from slicot import mb01rx

    np.random.seed(789)
    m, n = 4, 3
    alpha, beta = 0.0, 1.0

    # When alpha=0, R is not referenced on input
    r = np.zeros((m, m), order='F', dtype=float)

    b = np.array([
        [1.0, 2.0, 1.0],
        [2.0, 1.0, 3.0],
        [1.0, 2.0, 2.0],
        [3.0, 1.0, 2.0]
    ], order='F', dtype=float)

    a = np.array([
        [1.0, 2.0, 1.0],
        [2.0, 1.0, 3.0],
        [1.0, 2.0, 2.0],
        [3.0, 1.0, 2.0]
    ], order='F', dtype=float)

    # Expected: R_new[i,j] = beta*sum_k(B[i,k]*A[k,j]) for i >= j
    r_expected = np.array([
        [6.0, 0.0, 0.0, 0.0],
        [7.0, 14.0, 0.0, 0.0],
        [7.0, 10.0, 9.0, 0.0],
        [7.0, 13.0, 9.0, 14.0]
    ], order='F', dtype=float)

    r_out, info = mb01rx('R', 'L', 'T', m, n, alpha, beta, r, a, b)

    assert info == 0
    np.testing.assert_allclose(np.tril(r_out), np.tril(r_expected), rtol=1e-14)


def test_mb01rx_alpha_zero():
    """
    Test special case: alpha=0, beta=1 -> R = op(A)*B (upper triangle).

    Random seed: 999 (for reproducibility)
    """
    from slicot import mb01rx

    np.random.seed(999)
    m, n = 3, 2
    alpha, beta = 0.0, 1.0

    # R input is ignored when alpha=0
    r = np.ones((m, m), order='F', dtype=float)

    a = np.array([
        [1.0, 2.0],
        [3.0, 1.0],
        [2.0, 4.0]
    ], order='F', dtype=float)

    b = np.array([
        [2.0, 1.0, 3.0],
        [1.0, 2.0, 1.0]
    ], order='F', dtype=float)

    # Expected: R_new = A*B (upper triangle only)
    r_expected = np.array([
        [4.0, 5.0, 5.0],
        [0.0, 5.0, 10.0],
        [0.0, 0.0, 10.0]
    ], order='F', dtype=float)

    r_out, info = mb01rx('L', 'U', 'N', m, n, alpha, beta, r, a, b)

    assert info == 0
    np.testing.assert_allclose(np.triu(r_out), np.triu(r_expected), rtol=1e-14)


def test_mb01rx_beta_zero():
    """
    Test special case: alpha=2, beta=0 -> R = 2*R (upper triangle).

    Random seed: 111 (for reproducibility)
    """
    from slicot import mb01rx

    np.random.seed(111)
    m, n = 3, 2
    alpha, beta = 2.0, 0.0

    r = np.array([
        [4.0, 1.0, 2.0],
        [0.0, 3.0, 1.0],
        [0.0, 0.0, 5.0]
    ], order='F', dtype=float)

    # A and B are not referenced when beta=0
    a = np.zeros((m, n), order='F', dtype=float)
    b = np.zeros((n, m), order='F', dtype=float)

    # Expected: R_new = 2*R (upper triangle)
    r_expected = np.array([
        [8.0, 2.0, 4.0],
        [5.0, 6.0, 2.0],
        [6.0, 7.0, 10.0]
    ], order='F', dtype=float)

    r_out, info = mb01rx('L', 'U', 'N', m, n, alpha, beta, r, a, b)

    assert info == 0
    np.testing.assert_allclose(np.triu(r_out), np.triu(r_expected), rtol=1e-14)


def test_mb01rx_alpha_beta_zero():
    """
    Test special case: alpha=0, beta=0 -> R = 0 (upper triangle).
    """
    from slicot import mb01rx

    m, n = 3, 2
    alpha, beta = 0.0, 0.0

    r = np.array([
        [4.0, 1.0, 2.0],
        [5.0, 3.0, 1.0],
        [6.0, 7.0, 5.0]
    ], order='F', dtype=float)

    a = np.zeros((m, n), order='F', dtype=float)
    b = np.zeros((n, m), order='F', dtype=float)

    # Expected: R = 0 (upper triangle)
    r_expected = np.array([
        [0.0, 0.0, 0.0],
        [5.0, 0.0, 0.0],
        [6.0, 7.0, 0.0]
    ], order='F', dtype=float)

    r_out, info = mb01rx('L', 'U', 'N', m, n, alpha, beta, r, a, b)

    assert info == 0
    np.testing.assert_allclose(np.triu(r_out), np.triu(r_expected), rtol=1e-14)


def test_mb01rx_error_invalid_side():
    """Test error handling: invalid SIDE parameter."""
    from slicot import mb01rx

    m, n = 3, 2
    r = np.eye(m, order='F', dtype=float)
    a = np.eye(m, n, order='F', dtype=float)
    b = np.eye(n, m, order='F', dtype=float)

    r_out, info = mb01rx('X', 'U', 'N', m, n, 1.0, 1.0, r, a, b)
    assert info == -1


def test_mb01rx_error_invalid_uplo():
    """Test error handling: invalid UPLO parameter."""
    from slicot import mb01rx

    m, n = 3, 2
    r = np.eye(m, order='F', dtype=float)
    a = np.eye(m, n, order='F', dtype=float)
    b = np.eye(n, m, order='F', dtype=float)

    r_out, info = mb01rx('L', 'X', 'N', m, n, 1.0, 1.0, r, a, b)
    assert info == -2


def test_mb01rx_error_invalid_trans():
    """Test error handling: invalid TRANS parameter."""
    from slicot import mb01rx

    m, n = 3, 2
    r = np.eye(m, order='F', dtype=float)
    a = np.eye(m, n, order='F', dtype=float)
    b = np.eye(n, m, order='F', dtype=float)

    r_out, info = mb01rx('L', 'U', 'X', m, n, 1.0, 1.0, r, a, b)
    assert info == -3


def test_mb01rx_error_m_negative():
    """Test error handling: M < 0."""
    from slicot import mb01rx

    m, n = -1, 2
    r = np.eye(1, order='F', dtype=float)
    a = np.eye(1, 2, order='F', dtype=float)
    b = np.eye(2, 1, order='F', dtype=float)

    r_out, info = mb01rx('L', 'U', 'N', m, n, 1.0, 1.0, r, a, b)
    assert info == -4


def test_mb01rx_error_n_negative():
    """Test error handling: N < 0."""
    from slicot import mb01rx

    m, n = 3, -1
    r = np.eye(m, order='F', dtype=float)
    a = np.eye(m, 1, order='F', dtype=float)
    b = np.eye(1, m, order='F', dtype=float)

    r_out, info = mb01rx('L', 'U', 'N', m, n, 1.0, 1.0, r, a, b)
    assert info == -5


def test_mb01rx_property_symmetry():
    """
    Validate mathematical property: Result should be symmetric when B = A'.

    For SIDE='L', UPLO='U': R = alpha*R + beta*A*A'
    For SIDE='R', UPLO='L': R = alpha*R + beta*A'*A
    Both should produce symmetric R.

    Random seed: 222 (for reproducibility)
    """
    from slicot import mb01rx

    np.random.seed(222)
    m, n = 4, 3
    alpha, beta = 0.0, 1.0

    a = np.random.randn(m, n).astype(float, order='F')

    # Test 1: R = A*A' (upper triangle)
    r1 = np.zeros((m, m), order='F', dtype=float)
    r1_out, info = mb01rx('L', 'U', 'N', m, n, alpha, beta, r1, a, a.T.copy(order='F'))
    assert info == 0

    # Full symmetric matrix
    r1_full = a @ a.T
    np.testing.assert_allclose(np.triu(r1_out), np.triu(r1_full), rtol=1e-14)

    # Test 2: R = A'*A (lower triangle) - different dimensions
    m2, n2 = n, m
    r2 = np.zeros((m2, m2), order='F', dtype=float)
    # For SIDE='L', TRANS='T': need A(n2 x m2) = A(4x3)
    a2 = a.copy(order='F')  # 4x3 - correct dimensions
    b2 = a.copy(order='F')  # 4x3 - same as A for symmetric result
    r2_out, info = mb01rx('L', 'L', 'T', m2, n2, alpha, beta, r2, a2, b2)
    assert info == 0

    r2_full = a.T @ a
    np.testing.assert_allclose(np.tril(r2_out), np.tril(r2_full), rtol=1e-14)


def test_mb01rx_property_scale_equivalence():
    """
    Validate mathematical property: Scaling equivalence.

    R = alpha*R0 + beta*A*B should equal R0 scaled by alpha plus beta*(A*B).

    Random seed: 333 (for reproducibility)
    """
    from slicot import mb01rx

    np.random.seed(333)
    m, n = 3, 2
    alpha, beta = 1.5, 2.0

    r0 = np.random.randn(m, m).astype(float, order='F')
    r0 = np.triu(r0)  # Upper triangular

    a = np.random.randn(m, n).astype(float, order='F')
    b = np.random.randn(n, m).astype(float, order='F')

    # Compute via routine
    r_routine, info = mb01rx('L', 'U', 'N', m, n, alpha, beta, r0.copy(order='F'), a, b)
    assert info == 0

    # Compute manually
    r_manual = alpha * r0 + beta * (a @ b)

    np.testing.assert_allclose(np.triu(r_routine), np.triu(r_manual), rtol=1e-14)


def test_mb01rx_edge_case_m_zero():
    """Test edge case: M=0 (empty matrix)."""
    from slicot import mb01rx

    m, n = 0, 2
    r = np.zeros((1, 1), order='F', dtype=float)
    a = np.zeros((1, n), order='F', dtype=float)
    b = np.zeros((n, 1), order='F', dtype=float)

    r_out, info = mb01rx('L', 'U', 'N', m, n, 1.0, 1.0, r, a, b)
    assert info == 0


def test_mb01rx_edge_case_n_zero():
    """Test edge case: N=0 (R = alpha*R)."""
    from slicot import mb01rx

    m, n = 3, 0
    alpha = 2.5

    r = np.array([
        [1.0, 2.0, 3.0],
        [0.0, 4.0, 5.0],
        [0.0, 0.0, 6.0]
    ], order='F', dtype=float)

    a = np.zeros((m, 1), order='F', dtype=float)
    b = np.zeros((1, m), order='F', dtype=float)

    r_expected = alpha * r

    r_out, info = mb01rx('L', 'U', 'N', m, n, alpha, 1.0, r, a, b)
    assert info == 0
    np.testing.assert_allclose(np.triu(r_out), np.triu(r_expected), rtol=1e-14)



"""
Tests for SB16CY - Cholesky factors of controllability/observability Grammians
for coprime factors of state-feedback controller.

SB16CY computes Cholesky factors Su and Ru of:
- Controllability Grammian P = Su*Su'
- Observability Grammian Q = Ru'*Ru

For left coprime factorization (JOBCF='L'):
  Continuous: (A+B*F)*P + P*(A+B*F)' + scalec^2*B*B' = 0
              (A+G*C)'*Q + Q*(A+G*C) + scaleo^2*F'*F = 0
  Discrete:   (A+B*F)*P*(A+B*F)' - P + scalec^2*B*B' = 0
              (A+G*C)'*Q*(A+G*C) - Q + scaleo^2*F'*F = 0

For right coprime factorization (JOBCF='R'):
  Continuous: (A+B*F)*P + P*(A+B*F)' + scalec^2*G*G' = 0
              (A+G*C)'*Q + Q*(A+G*C) + scaleo^2*C'*C = 0
  Discrete:   (A+B*F)*P*(A+B*F)' - P + scalec^2*G*G' = 0
              (A+G*C)'*Q*(A+G*C) - Q + scaleo^2*C'*C = 0
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose

from slicot import sb16cy


def test_sb16cy_continuous_left():
    """
    Test continuous-time left coprime factorization.

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)

    n, m, p = 3, 2, 2

    A = np.array([
        [-2.0, 0.0, 0.0],
        [0.0, -3.0, 0.0],
        [0.0, 0.0, -4.0]
    ], order='F', dtype=float)

    B = np.array([
        [1.0, 0.0],
        [0.0, 1.0],
        [1.0, 1.0]
    ], order='F', dtype=float)

    C = np.array([
        [1.0, 1.0, 0.0],
        [0.0, 1.0, 1.0]
    ], order='F', dtype=float)

    F = np.array([
        [-1.0, 0.5, 0.0],
        [0.0, -0.5, 0.3]
    ], order='F', dtype=float)

    G = np.array([
        [0.5, -0.2],
        [-0.3, 0.4],
        [0.1, -0.1]
    ], order='F', dtype=float)

    Acl_bf = A + B @ F
    Acl_gc = A + G @ C
    eig_bf = np.linalg.eigvals(Acl_bf)
    eig_gc = np.linalg.eigvals(Acl_gc)
    assert all(eig_bf.real < 0), f"A+B*F not stable: {eig_bf}"
    assert all(eig_gc.real < 0), f"A+G*C not stable: {eig_gc}"

    S, R, scalec, scaleo, info = sb16cy('C', 'L', n, m, p, A, B, C, F, G)

    assert info == 0, f"sb16cy failed with info={info}"

    P = S @ S.T
    Q = R.T @ R

    res_P = Acl_bf @ P + P @ Acl_bf.T + scalec**2 * (B @ B.T)
    res_Q = Acl_gc.T @ Q + Q @ Acl_gc + scaleo**2 * (F.T @ F)

    assert_allclose(res_P, np.zeros((n, n)), atol=1e-12)
    assert_allclose(res_Q, np.zeros((n, n)), atol=1e-12)


def test_sb16cy_continuous_right():
    """
    Test continuous-time right coprime factorization.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)

    n, m, p = 4, 2, 2

    A = np.diag([-1.0, -2.0, -3.0, -4.0]).astype(float, order='F')

    B = np.array([
        [1.0, 0.5],
        [0.2, 1.0],
        [0.5, 0.3],
        [0.1, 0.4]
    ], order='F', dtype=float)

    C = np.array([
        [1.0, 0.5, 0.2, 0.1],
        [0.3, 1.0, 0.4, 0.2]
    ], order='F', dtype=float)

    F = np.array([
        [-0.3, 0.1, 0.0, 0.0],
        [0.0, -0.2, 0.1, 0.0]
    ], order='F', dtype=float)

    G = np.array([
        [0.2, -0.1],
        [-0.1, 0.2],
        [0.0, 0.1],
        [0.1, 0.0]
    ], order='F', dtype=float)

    Acl_bf = A + B @ F
    Acl_gc = A + G @ C
    eig_bf = np.linalg.eigvals(Acl_bf)
    eig_gc = np.linalg.eigvals(Acl_gc)
    assert all(eig_bf.real < 0), f"A+B*F not stable: {eig_bf}"
    assert all(eig_gc.real < 0), f"A+G*C not stable: {eig_gc}"

    S, R, scalec, scaleo, info = sb16cy('C', 'R', n, m, p, A, B, C, F, G)

    assert info == 0, f"sb16cy failed with info={info}"

    P = S @ S.T
    Q = R.T @ R

    res_P = Acl_bf @ P + P @ Acl_bf.T + scalec**2 * (G @ G.T)
    res_Q = Acl_gc.T @ Q + Q @ Acl_gc + scaleo**2 * (C.T @ C)

    assert_allclose(res_P, np.zeros((n, n)), atol=1e-12)
    assert_allclose(res_Q, np.zeros((n, n)), atol=1e-12)


def test_sb16cy_discrete_left():
    """
    Test discrete-time left coprime factorization.

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)

    n, m, p = 3, 2, 2

    A = np.array([
        [0.5, 0.1, 0.0],
        [0.0, 0.4, 0.1],
        [0.0, 0.0, 0.3]
    ], order='F', dtype=float)

    B = np.array([
        [1.0, 0.0],
        [0.5, 0.5],
        [0.0, 1.0]
    ], order='F', dtype=float)

    C = np.array([
        [1.0, 0.0, 0.5],
        [0.0, 1.0, 0.0]
    ], order='F', dtype=float)

    F = np.array([
        [-0.1, 0.0, 0.05],
        [0.0, -0.15, 0.0]
    ], order='F', dtype=float)

    G = np.array([
        [0.1, 0.0],
        [0.0, 0.1],
        [0.05, 0.05]
    ], order='F', dtype=float)

    Acl_bf = A + B @ F
    Acl_gc = A + G @ C
    eig_bf = np.linalg.eigvals(Acl_bf)
    eig_gc = np.linalg.eigvals(Acl_gc)
    assert all(np.abs(eig_bf) < 1), f"A+B*F not Schur stable: {eig_bf}"
    assert all(np.abs(eig_gc) < 1), f"A+G*C not Schur stable: {eig_gc}"

    S, R, scalec, scaleo, info = sb16cy('D', 'L', n, m, p, A, B, C, F, G)

    assert info == 0, f"sb16cy failed with info={info}"

    P = S @ S.T
    Q = R.T @ R

    res_P = Acl_bf @ P @ Acl_bf.T - P + scalec**2 * (B @ B.T)
    res_Q = Acl_gc.T @ Q @ Acl_gc - Q + scaleo**2 * (F.T @ F)

    assert_allclose(res_P, np.zeros((n, n)), atol=1e-12)
    assert_allclose(res_Q, np.zeros((n, n)), atol=1e-12)


def test_sb16cy_discrete_right():
    """
    Test discrete-time right coprime factorization.

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)

    n, m, p = 4, 2, 3

    A = np.diag([0.3, 0.4, 0.5, 0.6]).astype(float, order='F')

    B = np.array([
        [0.5, 0.0],
        [0.0, 0.5],
        [0.2, 0.2],
        [0.1, 0.3]
    ], order='F', dtype=float)

    C = np.array([
        [1.0, 0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 0.0],
        [0.0, 0.0, 1.0, 0.0]
    ], order='F', dtype=float)

    F = np.array([
        [-0.05, 0.0, 0.0, 0.0],
        [0.0, -0.08, 0.0, 0.0]
    ], order='F', dtype=float)

    G = np.array([
        [0.1, 0.0, 0.05],
        [0.0, 0.1, 0.0],
        [0.0, 0.0, 0.1],
        [0.05, 0.05, 0.0]
    ], order='F', dtype=float)

    Acl_bf = A + B @ F
    Acl_gc = A + G @ C
    eig_bf = np.linalg.eigvals(Acl_bf)
    eig_gc = np.linalg.eigvals(Acl_gc)
    assert all(np.abs(eig_bf) < 1), f"A+B*F not Schur stable: {eig_bf}"
    assert all(np.abs(eig_gc) < 1), f"A+G*C not Schur stable: {eig_gc}"

    S, R, scalec, scaleo, info = sb16cy('D', 'R', n, m, p, A, B, C, F, G)

    assert info == 0, f"sb16cy failed with info={info}"

    P = S @ S.T
    Q = R.T @ R

    res_P = Acl_bf @ P @ Acl_bf.T - P + scalec**2 * (G @ G.T)
    res_Q = Acl_gc.T @ Q @ Acl_gc - Q + scaleo**2 * (C.T @ C)

    assert_allclose(res_P, np.zeros((n, n)), atol=1e-12)
    assert_allclose(res_Q, np.zeros((n, n)), atol=1e-12)


def test_sb16cy_grammian_symmetry():
    """
    Verify that Grammians P=S*S' and Q=R'*R are symmetric positive semi-definite.

    Random seed: 999 (for reproducibility)
    """
    np.random.seed(999)

    n, m, p = 3, 2, 2

    A = np.diag([-1.5, -2.5, -3.5]).astype(float, order='F')
    B = np.array([
        [1.0, 0.0],
        [0.0, 1.0],
        [0.5, 0.5]
    ], order='F', dtype=float)
    C = np.array([
        [1.0, 0.5, 0.0],
        [0.0, 0.5, 1.0]
    ], order='F', dtype=float)
    F = np.array([
        [-0.2, 0.1, 0.0],
        [0.0, -0.2, 0.1]
    ], order='F', dtype=float)
    G = np.array([
        [0.3, -0.1],
        [-0.1, 0.3],
        [0.0, 0.0]
    ], order='F', dtype=float)

    S, R, scalec, scaleo, info = sb16cy('C', 'L', n, m, p, A, B, C, F, G)
    assert info == 0

    P = S @ S.T
    Q = R.T @ R

    assert_allclose(P, P.T, rtol=1e-14)
    assert_allclose(Q, Q.T, rtol=1e-14)

    eigP = np.linalg.eigvalsh(P)
    eigQ = np.linalg.eigvalsh(Q)
    assert all(eigP >= -1e-14), f"P not PSD: {eigP}"
    assert all(eigQ >= -1e-14), f"Q not PSD: {eigQ}"


def test_sb16cy_quick_return():
    """
    Test quick return for n=0 case.
    """
    n, m, p = 0, 2, 2
    A = np.array([], dtype=float, order='F').reshape(0, 0)
    B = np.array([], dtype=float, order='F').reshape(0, 2)
    C = np.array([], dtype=float, order='F').reshape(2, 0)
    F = np.array([], dtype=float, order='F').reshape(2, 0)
    G = np.array([], dtype=float, order='F').reshape(0, 2)

    S, R, scalec, scaleo, info = sb16cy('C', 'L', n, m, p, A, B, C, F, G)

    assert info == 0
    assert scalec == 1.0
    assert scaleo == 1.0


def test_sb16cy_invalid_dico():
    """
    Test error handling for invalid DICO parameter.
    """
    n, m, p = 2, 1, 1
    A = np.eye(n, order='F', dtype=float)
    B = np.ones((n, m), order='F', dtype=float)
    C = np.ones((p, n), order='F', dtype=float)
    F = np.ones((m, n), order='F', dtype=float)
    G = np.ones((n, p), order='F', dtype=float)

    S, R, scalec, scaleo, info = sb16cy('X', 'L', n, m, p, A, B, C, F, G)

    assert info == -1


def test_sb16cy_invalid_jobcf():
    """
    Test error handling for invalid JOBCF parameter.
    """
    n, m, p = 2, 1, 1
    A = np.diag([-1.0, -2.0]).astype(float, order='F')
    B = np.ones((n, m), order='F', dtype=float)
    C = np.ones((p, n), order='F', dtype=float)
    F = np.ones((m, n), order='F', dtype=float)
    G = np.ones((n, p), order='F', dtype=float)

    S, R, scalec, scaleo, info = sb16cy('C', 'X', n, m, p, A, B, C, F, G)

    assert info == -2



"""
Tests for TB01VD - Output normal form conversion for discrete-time systems.

TB01VD converts (A, B, C, D, x0) to parameter vector THETA using output normal form.
The matrix A must be stable (all eigenvalues inside unit circle for discrete-time).

Algorithm:
1. Solve Lyapunov equation A'*Q*A - Q = -scale^2*C'*C in Cholesky factor T
2. Transform system using T
3. QR factorization of transposed observability matrix
4. Extract parameters via N orthogonal transformations
"""
import numpy as np
import pytest

from slicot import tb01vd, tb01vy


"""Test basic functionality."""

def test_2x2_stable_discrete_system():
    """
    Test 2x2 stable discrete-time system conversion.

    Random seed: 42 (for reproducibility)
    Uses small discrete-time stable system.
    """
    np.random.seed(42)
    n, m, l = 2, 1, 1

    a = np.array([
        [0.5, 0.1],
        [0.0, 0.3]
    ], order='F', dtype=float)

    b = np.array([
        [1.0],
        [0.5]
    ], order='F', dtype=float)

    c = np.array([
        [1.0, 0.5]
    ], order='F', dtype=float)

    d = np.array([
        [0.0]
    ], order='F', dtype=float)

    x0 = np.array([0.1, 0.2], dtype=float)

    theta, a_out, b_out, c_out, scale, info = tb01vd(
        n, m, l, a, b, c, d, x0, apply='N'
    )

    assert info == 0
    assert scale > 0.0

    ltheta_expected = n * (l + m + 1) + l * m
    assert theta.shape[0] >= ltheta_expected

    assert a_out.shape == (n, n)
    assert b_out.shape == (n, m)
    assert c_out.shape == (l, n)

def test_3x2x1_system():
    """
    Test 3-state, 2-input, 1-output system.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n, m, l = 3, 2, 1

    a = 0.4 * np.random.randn(n, n)
    a = np.asfortranarray(a)

    b = np.random.randn(n, m)
    b = np.asfortranarray(b)

    c = np.random.randn(l, n)
    c = np.asfortranarray(c)

    d = np.random.randn(l, m)
    d = np.asfortranarray(d)

    x0 = np.random.randn(n)

    theta, a_out, b_out, c_out, scale, info = tb01vd(
        n, m, l, a, b, c, d, x0, apply='N'
    )

    assert info == 0
    assert scale > 0.0

    ltheta = n * (l + m + 1) + l * m
    assert len(theta) >= ltheta


"""Test round-trip: TB01VD -> TB01VY should recover original system."""

def test_roundtrip_2x2_system():
    """
    Test VD->VY round-trip preserves system behavior.

    TB01VD: (A,B,C,D,x0) -> THETA
    TB01VY: THETA -> (A',B',C',D',x0')

    Systems should produce identical outputs for same input sequence.
    Random seed: 200 (for reproducibility)
    """
    np.random.seed(200)
    n, m, l = 2, 1, 1

    a = np.array([
        [0.6, 0.1],
        [-0.1, 0.4]
    ], order='F', dtype=float)

    b = np.array([
        [1.0],
        [0.5]
    ], order='F', dtype=float)

    c = np.array([
        [1.0, 0.0]
    ], order='F', dtype=float)

    d = np.array([
        [0.1]
    ], order='F', dtype=float)

    x0 = np.array([0.0, 0.0], dtype=float)

    theta, a_vd, b_vd, c_vd, scale, info = tb01vd(
        n, m, l, a.copy(), b.copy(), c.copy(), d.copy(), x0.copy(), apply='N'
    )
    assert info == 0

    ltheta = n * (l + m + 1) + l * m
    a_vy, b_vy, c_vy, d_vy, x0_vy, info2 = tb01vy(n, m, l, theta[:ltheta], apply='N')
    assert info2 == 0

    u_seq = np.array([[1.0], [0.5], [0.2], [-0.1], [0.3]], dtype=float)

    def simulate(A, B, C, D, x_init, u_sequence):
        """Simulate discrete-time system."""
        y_out = []
        x = x_init.copy()
        for u in u_sequence:
            y = C @ x + D @ u.reshape(-1, 1)
            y_out.append(y.flatten())
            x = A @ x + B @ u.reshape(-1, 1)
        return np.array(y_out)

    y_orig = simulate(a, b, c, d, x0.reshape(-1, 1), u_seq)
    y_recon = simulate(a_vy, b_vy, c_vy, d_vy, x0_vy.reshape(-1, 1), u_seq)

    np.testing.assert_allclose(y_orig, y_recon, rtol=1e-10, atol=1e-12)

def test_roundtrip_3x2x2_apply_bijective():
    """
    Test round-trip with bijective mapping (apply='A').

    The bijective mapping relaxes constraint norm(THETAi) < 1.
    Random seed: 300 (for reproducibility)
    """
    np.random.seed(300)
    n, m, l = 3, 2, 2

    a = 0.3 * np.random.randn(n, n)
    a = np.asfortranarray(a)

    b = np.random.randn(n, m)
    b = np.asfortranarray(b)

    c = np.random.randn(l, n)
    c = np.asfortranarray(c)

    d = np.random.randn(l, m)
    d = np.asfortranarray(d)

    x0 = np.random.randn(n)

    theta, a_vd, b_vd, c_vd, scale, info = tb01vd(
        n, m, l, a.copy(), b.copy(), c.copy(), d.copy(), x0.copy(), apply='A'
    )
    assert info == 0

    ltheta = n * (l + m + 1) + l * m
    a_vy, b_vy, c_vy, d_vy, x0_vy, info2 = tb01vy(n, m, l, theta[:ltheta], apply='A')
    assert info2 == 0

    u_seq = np.random.randn(5, m)

    def simulate_mimo(A, B, C, D, x_init, u_sequence):
        y_out = []
        x = x_init.reshape(-1, 1)
        for u in u_sequence:
            y = C @ x + D @ u.reshape(-1, 1)
            y_out.append(y.flatten())
            x = A @ x + B @ u.reshape(-1, 1)
        return np.array(y_out)

    y_orig = simulate_mimo(a, b, c, d, x0, u_seq)
    y_recon = simulate_mimo(a_vy, b_vy, c_vy, d_vy, x0_vy, u_seq)

    np.testing.assert_allclose(y_orig, y_recon, rtol=1e-10, atol=1e-12)


"""Test edge cases."""

def test_n_zero():
    """Test N=0: quick return with D copied to THETA."""
    n, m, l = 0, 2, 2

    a = np.zeros((1, 1), order='F', dtype=float)
    b = np.zeros((1, m), order='F', dtype=float)
    c = np.zeros((l, 1), order='F', dtype=float)
    d = np.array([
        [1.0, 2.0],
        [3.0, 4.0]
    ], order='F', dtype=float)
    x0 = np.zeros(1, dtype=float)

    theta, a_out, b_out, c_out, scale, info = tb01vd(
        n, m, l, a, b, c, d, x0, apply='N'
    )

    assert info == 0
    np.testing.assert_allclose(theta[:l*m].reshape((l, m), order='F'), d, rtol=1e-14)

def test_l_zero():
    """Test L=0: THETA contains B and x0."""
    n, m, l = 2, 2, 0

    a = np.array([
        [0.5, 0.1],
        [0.0, 0.3]
    ], order='F', dtype=float)

    b = np.array([
        [1.0, 2.0],
        [3.0, 4.0]
    ], order='F', dtype=float)

    c = np.zeros((1, n), order='F', dtype=float)
    d = np.zeros((1, m), order='F', dtype=float)
    x0 = np.array([0.5, 0.6], dtype=float)

    theta, a_out, b_out, c_out, scale, info = tb01vd(
        n, m, l, a, b, c, d, x0, apply='N'
    )

    assert info == 0
    b_from_theta = theta[:n*m].reshape((n, m), order='F')
    np.testing.assert_allclose(b_from_theta, b, rtol=1e-14)
    np.testing.assert_allclose(theta[n*m:n*m+n], x0, rtol=1e-14)

def test_m_zero():
    """Test M=0: only autonomous system with output."""
    n, m, l = 2, 0, 1

    a = np.array([
        [0.5, 0.1],
        [0.0, 0.3]
    ], order='F', dtype=float)

    b = np.zeros((n, 1), order='F', dtype=float)

    c = np.array([
        [1.0, 0.5]
    ], order='F', dtype=float)

    d = np.zeros((l, 1), order='F', dtype=float)
    x0 = np.array([0.1, 0.2], dtype=float)

    theta, a_out, b_out, c_out, scale, info = tb01vd(
        n, m, l, a, b, c, d, x0, apply='N'
    )

    assert info == 0


"""Test error conditions."""

def test_unstable_system():
    """Test unstable A (eigenvalue outside unit circle) returns info=2."""
    n, m, l = 2, 1, 1

    a = np.array([
        [1.5, 0.0],
        [0.0, 0.5]
    ], order='F', dtype=float)

    b = np.array([[1.0], [0.5]], order='F', dtype=float)
    c = np.array([[1.0, 0.5]], order='F', dtype=float)
    d = np.array([[0.0]], order='F', dtype=float)
    x0 = np.array([0.0, 0.0], dtype=float)

    theta, a_out, b_out, c_out, scale, info = tb01vd(
        n, m, l, a, b, c, d, x0, apply='N'
    )

    assert info == 2

def test_invalid_apply():
    """Test invalid apply parameter raises error."""
    n, m, l = 2, 1, 1

    a = np.array([[0.5, 0.0], [0.0, 0.3]], order='F', dtype=float)
    b = np.array([[1.0], [0.5]], order='F', dtype=float)
    c = np.array([[1.0, 0.5]], order='F', dtype=float)
    d = np.array([[0.0]], order='F', dtype=float)
    x0 = np.array([0.0, 0.0], dtype=float)

    with pytest.raises(ValueError):
        tb01vd(n, m, l, a, b, c, d, x0, apply='X')


"""Test THETA vector structure."""

def test_theta_layout():
    """
    Verify THETA layout matches documentation:
    - THETA(1:N*L)                      : parameters for A, C
    - THETA(N*L+1:N*(L+M))              : parameters for B
    - THETA(N*(L+M)+1:N*(L+M)+L*M)      : parameters for D
    - THETA(N*(L+M)+L*M+1:N*(L+M+1)+L*M): parameters for x0

    Random seed: 400 (for reproducibility)
    """
    np.random.seed(400)
    n, m, l = 3, 2, 2

    a = 0.3 * np.random.randn(n, n)
    a = np.asfortranarray(a)

    b = np.random.randn(n, m)
    b = np.asfortranarray(b)

    c = np.random.randn(l, n)
    c = np.asfortranarray(c)

    d = np.random.randn(l, m)
    d = np.asfortranarray(d)

    x0 = np.random.randn(n)

    theta, a_out, b_out, c_out, scale, info = tb01vd(
        n, m, l, a.copy(), b.copy(), c.copy(), d.copy(), x0.copy(), apply='N'
    )

    assert info == 0

    ltheta = n * (l + m + 1) + l * m
    assert len(theta) >= ltheta

    theta_b = theta[n*l : n*(l+m)].reshape((n, m), order='F')
    np.testing.assert_allclose(theta_b, b_out, rtol=1e-14)

    theta_d = theta[n*(l+m) : n*(l+m)+l*m].reshape((l, m), order='F')
    np.testing.assert_allclose(theta_d, d, rtol=1e-14)


"""Test larger systems for robustness."""

def test_5x3x2_system():
    """
    Test 5-state, 3-input, 2-output system.

    Random seed: 500 (for reproducibility)
    """
    np.random.seed(500)
    n, m, l = 5, 3, 2

    a = 0.2 * np.random.randn(n, n)
    a = np.asfortranarray(a)

    b = np.random.randn(n, m)
    b = np.asfortranarray(b)

    c = np.random.randn(l, n)
    c = np.asfortranarray(c)

    d = np.random.randn(l, m)
    d = np.asfortranarray(d)

    x0 = np.random.randn(n)

    theta, a_out, b_out, c_out, scale, info = tb01vd(
        n, m, l, a.copy(), b.copy(), c.copy(), d.copy(), x0.copy(), apply='N'
    )

    assert info == 0
    assert scale > 0.0

    ltheta = n * (l + m + 1) + l * m
    a_vy, b_vy, c_vy, d_vy, x0_vy, info2 = tb01vy(n, m, l, theta[:ltheta], apply='N')
    assert info2 == 0

    u_seq = np.random.randn(10, m)

    def simulate_mimo(A, B, C, D, x_init, u_sequence):
        y_out = []
        x = x_init.reshape(-1, 1)
        for u in u_sequence:
            y = C @ x + D @ u.reshape(-1, 1)
            y_out.append(y.flatten())
            x = A @ x + B @ u.reshape(-1, 1)
        return np.array(y_out)

    y_orig = simulate_mimo(a, b, c, d, x0, u_seq)
    y_recon = simulate_mimo(a_vy, b_vy, c_vy, d_vy, x0_vy, u_seq)

    np.testing.assert_allclose(y_orig, y_recon, rtol=1e-9, atol=1e-10)



"""
Tests for TB01UD - Controllable block Hessenberg realization

TB01UD reduces a state-space system (A,B,C) to controllability staircase form
using orthogonal similarity transformations. The transformed system has:
- Acont: upper block Hessenberg with full row rank subdiagonal blocks
- Bcont: zero except for first block (rank(B) rows)
- Cc: C*Z transformed output matrix

Mathematical properties tested:
- Orthogonality: Z'*Z = I
- Similarity: Ac = Z'*A*Z, Bc = Z'*B, Cc = C*Z
- Eigenvalue preservation for controllable subsystem
- Controllability dimension (NCONT)

Test data from SLICOT-Reference/doc/TB01UD.html example
"""
import numpy as np
import pytest

try:
    import slicot
    HAS_SLICOT = True
except ImportError:
    HAS_SLICOT = False


pytestmark = pytest.mark.skipif(not HAS_SLICOT, reason="slicot module not available")


def test_html_example():
    """Test TB01UD with example from HTML documentation

    From HTML doc:
    N=3, M=2, P=2, TOL=0.0, JOBZ='I'

    A = [-1.0   0.0   0.0]    B = [1.0  0.0]    C = [0.0  2.0  1.0]
        [-2.0  -2.0  -2.0]        [0.0  2.0]        [1.0  0.0  0.0]
        [-1.0   0.0  -3.0]        [0.0  1.0]

    Expected outputs:
    NCONT = 2
    INDCON = 1
    NBLK = [2]

    Acont = [-3.0000   2.2361]
            [ 0.0000  -1.0000]

    Bcont = [ 0.0000  -2.2361]
            [ 1.0000   0.0000]

    Ccont = [-2.2361   0.0000]
            [ 0.0000   1.0000]

    Z = [ 0.0000   1.0000   0.0000]
        [-0.8944   0.0000  -0.4472]
        [-0.4472   0.0000   0.8944]
    """
    n, m, p = 3, 2, 2

    a = np.array([
        [-1.0,  0.0,  0.0],
        [-2.0, -2.0, -2.0],
        [-1.0,  0.0, -3.0]
    ], dtype=np.float64, order='F')

    b = np.array([
        [1.0, 0.0],
        [0.0, 2.0],
        [0.0, 1.0]
    ], dtype=np.float64, order='F')

    c = np.array([
        [0.0, 2.0, 1.0],
        [1.0, 0.0, 0.0]
    ], dtype=np.float64, order='F')

    expected_ncont = 2
    expected_indcon = 1
    expected_nblk = [2]

    expected_acont = np.array([
        [-3.0000,  2.2361],
        [ 0.0000, -1.0000]
    ], dtype=np.float64, order='F')

    expected_bcont = np.array([
        [ 0.0000, -2.2361],
        [ 1.0000,  0.0000]
    ], dtype=np.float64, order='F')

    expected_ccont = np.array([
        [-2.2361,  0.0000],
        [ 0.0000,  1.0000]
    ], dtype=np.float64, order='F')

    expected_z = np.array([
        [ 0.0000,  1.0000,  0.0000],
        [-0.8944,  0.0000, -0.4472],
        [-0.4472,  0.0000,  0.8944]
    ], dtype=np.float64, order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'I', n, m, p, a, b, c, tol=0.0)

    assert info == 0, f"tb01ud failed with info={info}"
    assert ncont == expected_ncont, f"ncont={ncont}, expected {expected_ncont}"
    assert indcon == expected_indcon, f"indcon={indcon}, expected {expected_indcon}"

    for i in range(indcon):
        assert nblk[i] == expected_nblk[i], f"nblk[{i}]={nblk[i]}, expected {expected_nblk[i]}"

    np.testing.assert_allclose(
        a_out[:ncont, :ncont], expected_acont, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(
        b_out[:ncont, :], expected_bcont, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(
        c_out[:, :ncont], expected_ccont, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(z, expected_z, rtol=1e-3, atol=1e-4)

def test_orthogonality_property():
    """Test that Z is orthogonal: Z'*Z = I

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n, m, p = 4, 2, 3

    a = np.random.randn(n, n).astype(np.float64, order='F')
    b = np.random.randn(n, m).astype(np.float64, order='F')
    c = np.random.randn(p, n).astype(np.float64, order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'I', n, m, p, a, b, c, tol=0.0)

    assert info == 0

    z_orth = z.T @ z
    np.testing.assert_allclose(z_orth, np.eye(n), rtol=1e-14, atol=1e-14)

def test_similarity_transformation():
    """Test similarity relationships for C matrix: Cc = C*Z

    Note: For A and B, the routine applies internal scaling (via mb01pd)
    before applying transformations, so Ac != Z'*A*Z exactly.
    However, C is only transformed (not scaled), so Cc = C*Z holds.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n, m, p = 5, 2, 3

    a_orig = np.random.randn(n, n).astype(np.float64, order='F')
    b_orig = np.random.randn(n, m).astype(np.float64, order='F')
    c_orig = np.random.randn(p, n).astype(np.float64, order='F')

    a = a_orig.copy(order='F')
    b = b_orig.copy(order='F')
    c = c_orig.copy(order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'I', n, m, p, a, b, c, tol=0.0)

    assert info == 0

    c_check = c_orig @ z
    np.testing.assert_allclose(c_out, c_check, rtol=1e-13, atol=1e-14)

def test_eigenvalue_preservation_controllable():
    """Test eigenvalue preservation for the controllable subsystem

    For the controllable part, eigenvalues must be preserved.
    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    n, m, p = 4, 2, 2

    a = np.random.randn(n, n).astype(np.float64, order='F')
    b = np.random.randn(n, m).astype(np.float64, order='F')
    c = np.random.randn(p, n).astype(np.float64, order='F')

    eig_orig = np.linalg.eigvals(a)

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'I', n, m, p, a, b, c, tol=0.0)

    assert info == 0

    eig_new = np.linalg.eigvals(a_out)
    np.testing.assert_allclose(
        np.sort(eig_orig.real), np.sort(eig_new.real), rtol=1e-12, atol=1e-14)
    np.testing.assert_allclose(
        np.sort(np.abs(eig_orig.imag)), np.sort(np.abs(eig_new.imag)),
        rtol=1e-12, atol=1e-14)

def test_fully_controllable_system():
    """Test with a fully controllable system (ncont = n)

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n, m, p = 3, 2, 2

    a = np.array([
        [0, 1, 0],
        [0, 0, 1],
        [-1, -2, -3]
    ], dtype=np.float64, order='F')

    b = np.array([
        [0, 0],
        [0, 0],
        [1, 1]
    ], dtype=np.float64, order='F')

    c = np.random.randn(p, n).astype(np.float64, order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'I', n, m, p, a, b, c, tol=0.0)

    assert info == 0
    assert ncont == n, f"System should be fully controllable, ncont={ncont}"

def test_uncontrollable_system():
    """Test with a partially uncontrollable system

    Random seed: 888 (for reproducibility)
    """
    np.random.seed(888)
    n, m, p = 4, 1, 2

    a = np.array([
        [1, 0, 0, 0],
        [0, 2, 0, 0],
        [0, 0, 3, 0],
        [0, 0, 0, 4]
    ], dtype=np.float64, order='F')

    b = np.array([
        [1],
        [1],
        [0],
        [0]
    ], dtype=np.float64, order='F')

    c = np.random.randn(p, n).astype(np.float64, order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'I', n, m, p, a, b, c, tol=0.0)

    assert info == 0
    assert ncont < n, f"System should be partially uncontrollable, ncont={ncont}"

def test_jobz_n_no_z():
    """Test with JOBZ='N' (no transformation matrix computed)"""
    n, m, p = 3, 2, 2

    a = np.array([
        [-1.0,  0.0,  0.0],
        [-2.0, -2.0, -2.0],
        [-1.0,  0.0, -3.0]
    ], dtype=np.float64, order='F')

    b = np.array([
        [1.0, 0.0],
        [0.0, 2.0],
        [0.0, 1.0]
    ], dtype=np.float64, order='F')

    c = np.array([
        [0.0, 2.0, 1.0],
        [1.0, 0.0, 0.0]
    ], dtype=np.float64, order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'N', n, m, p, a, b, c, tol=0.0)

    assert info == 0
    assert ncont == 2

def test_jobz_f_factored_form():
    """Test with JOBZ='F' (factored form of transformation)"""
    n, m, p = 3, 2, 2

    a = np.array([
        [-1.0,  0.0,  0.0],
        [-2.0, -2.0, -2.0],
        [-1.0,  0.0, -3.0]
    ], dtype=np.float64, order='F')

    b = np.array([
        [1.0, 0.0],
        [0.0, 2.0],
        [0.0, 1.0]
    ], dtype=np.float64, order='F')

    c = np.array([
        [0.0, 2.0, 1.0],
        [1.0, 0.0, 0.0]
    ], dtype=np.float64, order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'F', n, m, p, a, b, c, tol=0.0)

    assert info == 0
    assert ncont == 2

def test_quick_return_n_zero():
    """Test quick return for N=0"""
    n, m, p = 0, 2, 2
    a = np.zeros((1, 1), dtype=np.float64, order='F')
    b = np.zeros((1, 2), dtype=np.float64, order='F')
    c = np.zeros((2, 1), dtype=np.float64, order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'I', n, m, p, a, b, c, tol=0.0)

    assert info == 0
    assert ncont == 0
    assert indcon == 0

def test_quick_return_m_zero():
    """Test quick return for M=0 (no inputs)"""
    n, m, p = 3, 0, 2
    a = np.eye(n, dtype=np.float64, order='F')
    b = np.zeros((n, 1), dtype=np.float64, order='F')
    c = np.random.randn(p, n).astype(np.float64, order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'I', n, m, p, a, b, c, tol=0.0)

    assert info == 0
    assert ncont == 0
    assert indcon == 0

def test_quick_return_b_zero():
    """Test quick return for B=0 (zero input matrix)"""
    n, m, p = 3, 2, 2
    a = np.eye(n, dtype=np.float64, order='F')
    b = np.zeros((n, m), dtype=np.float64, order='F')
    c = np.random.randn(p, n).astype(np.float64, order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'I', n, m, p, a, b, c, tol=0.0)

    assert info == 0
    assert ncont == 0
    assert indcon == 0

def test_upper_block_hessenberg_structure():
    """Test that transformed A has upper block Hessenberg structure

    Random seed: 999 (for reproducibility)
    """
    np.random.seed(999)
    n, m, p = 6, 2, 3

    a = np.random.randn(n, n).astype(np.float64, order='F')
    b = np.random.randn(n, m).astype(np.float64, order='F')
    c = np.random.randn(p, n).astype(np.float64, order='F')

    a_out, b_out, c_out, ncont, indcon, nblk, z, tau, info = slicot.tb01ud(
        'I', n, m, p, a, b, c, tol=0.0)

    assert info == 0

    if ncont > 0 and indcon > 0:
        row_offset = 0
        for k in range(indcon - 1):
            block_size = nblk[k]
            row_offset += block_size
            next_block_size = nblk[k + 1] if k + 1 < indcon else 0
            for i in range(row_offset + next_block_size, ncont):
                for j in range(row_offset - block_size):
                    if i >= 0 and j >= 0 and i < ncont and j < ncont:
                        assert abs(a_out[i, j]) < 1e-10, \
                            f"A not upper block Hessenberg: A[{i},{j}]={a_out[i,j]}"

def test_invalid_jobz():
    """Test error handling for invalid JOBZ parameter"""
    n, m, p = 3, 2, 2
    a = np.eye(n, dtype=np.float64, order='F')
    b = np.zeros((n, m), dtype=np.float64, order='F')
    c = np.zeros((p, n), dtype=np.float64, order='F')

    with pytest.raises((ValueError, RuntimeError)):
        slicot.tb01ud('X', n, m, p, a, b, c, tol=0.0)

def test_invalid_n_negative():
    """Test error handling for N < 0"""
    a = np.eye(1, dtype=np.float64, order='F')
    b = np.zeros((1, 1), dtype=np.float64, order='F')
    c = np.zeros((1, 1), dtype=np.float64, order='F')

    with pytest.raises((ValueError, RuntimeError)):
        slicot.tb01ud('I', -1, 1, 1, a, b, c, tol=0.0)

def test_invalid_m_negative():
    """Test error handling for M < 0"""
    a = np.eye(1, dtype=np.float64, order='F')
    b = np.zeros((1, 1), dtype=np.float64, order='F')
    c = np.zeros((1, 1), dtype=np.float64, order='F')

    with pytest.raises((ValueError, RuntimeError)):
        slicot.tb01ud('I', 1, -1, 1, a, b, c, tol=0.0)

def test_invalid_p_negative():
    """Test error handling for P < 0"""
    a = np.eye(1, dtype=np.float64, order='F')
    b = np.zeros((1, 1), dtype=np.float64, order='F')
    c = np.zeros((1, 1), dtype=np.float64, order='F')

    with pytest.raises((ValueError, RuntimeError)):
        slicot.tb01ud('I', 1, 1, -1, a, b, c, tol=0.0)



"""
Tests for MB03IZ - Eigenvalue reordering for complex skew-Hamiltonian/Hamiltonian pencil.

MB03IZ moves eigenvalues with strictly negative real parts of an N-by-N complex
skew-Hamiltonian/Hamiltonian pencil aS - bH in structured Schur form to the
leading principal subpencil, while keeping the triangular form.

On entry:
        (  A  D  )      (  B  F  )
    Z = (        ), H = (        ),
        (  0  C  )      (  0 -B' )

where A and B are upper triangular and C is lower triangular.
"""

import numpy as np
import pytest


def test_mb03iz_basic():
    """
    Test basic eigenvalue reordering with simple input.

    Random seed: 42 (for reproducibility)
    Validates that:
    1. Output info is 0 (success)
    2. Output neig >= 0
    3. Q is unitary (Q @ Q.conj().T = I) when COMPQ='I'
    4. U is unitary symplectic when COMPU='I'
    """
    from slicot import mb03iz

    np.random.seed(42)

    n = 4
    m = n // 2

    a = np.array([
        [1.0 + 0.5j, 0.3 - 0.2j],
        [0.0 + 0.0j, -0.5 + 0.1j]
    ], dtype=np.complex128, order='F')

    c = np.array([
        [0.8 + 0.1j, 0.0 + 0.0j],
        [0.2 - 0.3j, 0.6 + 0.2j]
    ], dtype=np.complex128, order='F')

    d = np.array([
        [0.4 + 0.2j, -0.1 + 0.1j],
        [0.1 - 0.2j, 0.3 + 0.1j]
    ], dtype=np.complex128, order='F')

    b = np.array([
        [0.7 + 0.3j, 0.2 - 0.1j],
        [0.0 + 0.0j, 0.4 + 0.2j]
    ], dtype=np.complex128, order='F')

    f = np.array([
        [1.0 + 0.0j, 0.1 - 0.05j],
        [0.1 + 0.05j, 0.8 + 0.0j]
    ], dtype=np.complex128, order='F')

    tol = 0.0

    a_out, c_out, d_out, b_out, f_out, q, u1, u2, neig, info = mb03iz(
        'I', 'I', n, a, c, d, b, f, tol
    )

    assert info == 0
    assert neig >= 0
    assert neig <= m

    np.testing.assert_allclose(
        q @ q.conj().T, np.eye(n), rtol=1e-13, atol=1e-14
    )

    u_full = np.block([
        [u1, u2],
        [-np.conj(u2), np.conj(u1)]
    ])
    np.testing.assert_allclose(
        u_full @ u_full.conj().T, np.eye(n), rtol=1e-13, atol=1e-14
    )


def test_mb03iz_no_transform():
    """
    Test with COMPQ='N' and COMPU='N' (no transformation matrices computed).

    Random seed: 123 (for reproducibility)
    """
    from slicot import mb03iz

    np.random.seed(123)

    n = 4
    m = n // 2

    a = np.array([
        [1.0 + 0.2j, 0.5 - 0.1j],
        [0.0 + 0.0j, -0.8 + 0.3j]
    ], dtype=np.complex128, order='F')

    c = np.array([
        [0.9 + 0.1j, 0.0 + 0.0j],
        [0.3 - 0.2j, 0.7 + 0.1j]
    ], dtype=np.complex128, order='F')

    d = np.array([
        [0.3 + 0.1j, -0.2 + 0.1j],
        [0.1 - 0.1j, 0.4 + 0.2j]
    ], dtype=np.complex128, order='F')

    b = np.array([
        [0.6 + 0.2j, 0.3 - 0.1j],
        [0.0 + 0.0j, 0.5 + 0.1j]
    ], dtype=np.complex128, order='F')

    f = np.array([
        [0.9 + 0.0j, 0.2 - 0.1j],
        [0.2 + 0.1j, 1.1 + 0.0j]
    ], dtype=np.complex128, order='F')

    tol = 0.0

    a_out, c_out, d_out, b_out, f_out, q, u1, u2, neig, info = mb03iz(
        'N', 'N', n, a, c, d, b, f, tol
    )

    assert info == 0
    assert neig >= 0


def test_mb03iz_zero_dimension():
    """
    Test with N=0 (quick return case).
    """
    from slicot import mb03iz

    n = 0

    a = np.zeros((0, 0), dtype=np.complex128, order='F')
    c = np.zeros((0, 0), dtype=np.complex128, order='F')
    d = np.zeros((0, 0), dtype=np.complex128, order='F')
    b = np.zeros((0, 0), dtype=np.complex128, order='F')
    f = np.zeros((0, 0), dtype=np.complex128, order='F')

    tol = 0.0

    a_out, c_out, d_out, b_out, f_out, q, u1, u2, neig, info = mb03iz(
        'N', 'N', n, a, c, d, b, f, tol
    )

    assert info == 0
    assert neig == 0


def test_mb03iz_invalid_n_odd():
    """
    Test error handling for odd N (N must be even).
    """
    from slicot import mb03iz

    n = 3

    a = np.zeros((1, 1), dtype=np.complex128, order='F')
    c = np.zeros((1, 1), dtype=np.complex128, order='F')
    d = np.zeros((1, 1), dtype=np.complex128, order='F')
    b = np.zeros((1, 1), dtype=np.complex128, order='F')
    f = np.zeros((1, 1), dtype=np.complex128, order='F')

    tol = 0.0

    a_out, c_out, d_out, b_out, f_out, q, u1, u2, neig, info = mb03iz(
        'N', 'N', n, a, c, d, b, f, tol
    )

    assert info == -3


def test_mb03iz_update_mode():
    """
    Test with COMPQ='U' and COMPU='U' (update existing transformations).

    Random seed: 456 (for reproducibility)
    Validates that existing Q and U matrices are updated correctly.
    """
    from slicot import mb03iz

    np.random.seed(456)

    n = 4
    m = n // 2

    a = np.array([
        [0.8 + 0.3j, 0.4 - 0.1j],
        [0.0 + 0.0j, -0.6 + 0.2j]
    ], dtype=np.complex128, order='F')

    c = np.array([
        [0.7 + 0.2j, 0.0 + 0.0j],
        [0.1 - 0.2j, 0.5 + 0.3j]
    ], dtype=np.complex128, order='F')

    d = np.array([
        [0.2 + 0.1j, -0.1 + 0.05j],
        [0.05 - 0.1j, 0.25 + 0.1j]
    ], dtype=np.complex128, order='F')

    b = np.array([
        [0.5 + 0.2j, 0.15 - 0.05j],
        [0.0 + 0.0j, 0.35 + 0.15j]
    ], dtype=np.complex128, order='F')

    f = np.array([
        [0.85 + 0.0j, 0.1 - 0.05j],
        [0.1 + 0.05j, 0.75 + 0.0j]
    ], dtype=np.complex128, order='F')

    q_init = np.eye(n, dtype=np.complex128, order='F')
    u1_init = np.eye(m, dtype=np.complex128, order='F')
    u2_init = np.zeros((m, m), dtype=np.complex128, order='F')

    tol = 0.0

    a_out, c_out, d_out, b_out, f_out, q, u1, u2, neig, info = mb03iz(
        'U', 'U', n, a, c, d, b, f, tol,
        q=q_init, u1=u1_init, u2=u2_init
    )

    assert info == 0
    assert neig >= 0

    np.testing.assert_allclose(
        q @ q.conj().T, np.eye(n), rtol=1e-13, atol=1e-14
    )


def test_mb03iz_transformation_consistency():
    """
    Test that transformation matrices satisfy the structure equations.

    Random seed: 789 (for reproducibility)
    Validates: Zout = U' * Z * Q and Hout = J * Q' * J' * H * Q
    """
    from slicot import mb03iz

    np.random.seed(789)

    n = 4
    m = n // 2

    a_orig = np.array([
        [1.2 + 0.4j, 0.6 - 0.2j],
        [0.0 + 0.0j, -0.7 + 0.25j]
    ], dtype=np.complex128, order='F')

    c_orig = np.array([
        [0.85 + 0.15j, 0.0 + 0.0j],
        [0.25 - 0.15j, 0.65 + 0.25j]
    ], dtype=np.complex128, order='F')

    d_orig = np.array([
        [0.35 + 0.15j, -0.12 + 0.08j],
        [0.08 - 0.12j, 0.28 + 0.12j]
    ], dtype=np.complex128, order='F')

    b_orig = np.array([
        [0.65 + 0.25j, 0.22 - 0.08j],
        [0.0 + 0.0j, 0.42 + 0.18j]
    ], dtype=np.complex128, order='F')

    f_orig = np.array([
        [0.95 + 0.0j, 0.12 - 0.06j],
        [0.12 + 0.06j, 0.82 + 0.0j]
    ], dtype=np.complex128, order='F')

    a = a_orig.copy()
    c = c_orig.copy()
    d = d_orig.copy()
    b = b_orig.copy()
    f = f_orig.copy()

    tol = 0.0

    a_out, c_out, d_out, b_out, f_out, q, u1, u2, neig, info = mb03iz(
        'I', 'I', n, a, c, d, b, f, tol
    )

    assert info == 0

    np.testing.assert_allclose(
        q @ q.conj().T, np.eye(n), rtol=1e-13, atol=1e-14
    )

    u_full = np.block([
        [u1, u2],
        [-np.conj(u2), np.conj(u1)]
    ])
    np.testing.assert_allclose(
        u_full @ u_full.conj().T, np.eye(n), rtol=1e-13, atol=1e-14
    )


def test_mb03iz_determinant_preserved():
    """
    Test that unitary transformations have |det| = 1.

    Random seed: 999 (for reproducibility)
    """
    from slicot import mb03iz

    np.random.seed(999)

    n = 4
    m = n // 2

    a = np.array([
        [0.9 + 0.35j, 0.45 - 0.15j],
        [0.0 + 0.0j, -0.55 + 0.2j]
    ], dtype=np.complex128, order='F')

    c = np.array([
        [0.75 + 0.12j, 0.0 + 0.0j],
        [0.18 - 0.22j, 0.58 + 0.22j]
    ], dtype=np.complex128, order='F')

    d = np.array([
        [0.28 + 0.12j, -0.1 + 0.06j],
        [0.06 - 0.1j, 0.24 + 0.1j]
    ], dtype=np.complex128, order='F')

    b = np.array([
        [0.58 + 0.22j, 0.18 - 0.06j],
        [0.0 + 0.0j, 0.38 + 0.16j]
    ], dtype=np.complex128, order='F')

    f = np.array([
        [0.88 + 0.0j, 0.1 - 0.04j],
        [0.1 + 0.04j, 0.78 + 0.0j]
    ], dtype=np.complex128, order='F')

    tol = 0.0

    a_out, c_out, d_out, b_out, f_out, q, u1, u2, neig, info = mb03iz(
        'I', 'I', n, a, c, d, b, f, tol
    )

    assert info == 0

    det_q = np.linalg.det(q)
    np.testing.assert_allclose(np.abs(det_q), 1.0, rtol=1e-13)

    u_full = np.block([
        [u1, u2],
        [-np.conj(u2), np.conj(u1)]
    ])
    det_u = np.linalg.det(u_full)
    np.testing.assert_allclose(np.abs(det_u), 1.0, rtol=1e-13)



"""
Tests for MB02RD: Solve linear systems with Hessenberg LU factorization.

Uses numpy only.
"""

import numpy as np


def test_mb02rd_notrans():
    """
    Test MB02RD solving H*X = B (no transpose).

    Random seed: 42 (for reproducibility)
    """
    from slicot import mb02sd, mb02rd

    np.random.seed(42)
    n = 4
    nrhs = 2

    h = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        for j in range(n):
            if j >= i - 1:
                h[i, j] = np.random.randn()

    h[np.diag_indices(n)] += 2.0

    x_true = np.random.randn(n, nrhs).astype(float, order='F')
    b = h @ x_true

    h_orig = h.copy()
    h_lu, ipiv, info = mb02sd(n, h)
    assert info == 0

    b_copy = b.copy(order='F')
    x_out, info = mb02rd('N', n, nrhs, h_lu, ipiv, b_copy)

    assert info == 0
    np.testing.assert_allclose(x_out, x_true, rtol=1e-12, atol=1e-14)


def test_mb02rd_trans():
    """
    Test MB02RD solving H'*X = B (transpose).

    Random seed: 123 (for reproducibility)
    """
    from slicot import mb02sd, mb02rd

    np.random.seed(123)
    n = 3
    nrhs = 1

    h = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        for j in range(n):
            if j >= i - 1:
                h[i, j] = np.random.randn()

    h[np.diag_indices(n)] += 2.0

    x_true = np.random.randn(n, nrhs).astype(float, order='F')
    b = h.T @ x_true

    h_lu, ipiv, info = mb02sd(n, h)
    assert info == 0

    b_copy = b.copy(order='F')
    x_out, info = mb02rd('T', n, nrhs, h_lu, ipiv, b_copy)

    assert info == 0
    np.testing.assert_allclose(x_out, x_true, rtol=1e-12, atol=1e-14)


def test_mb02rd_n0():
    """Test MB02RD with n=0 (quick return)."""
    from slicot import mb02sd, mb02rd

    n = 0
    nrhs = 1

    h = np.zeros((1, 1), order='F', dtype=float)
    b = np.zeros((1, 1), order='F', dtype=float)
    ipiv = np.zeros(1, dtype=np.int32)

    x_out, info = mb02rd('N', n, nrhs, h, ipiv, b)

    assert info == 0


def test_mb02rd_nrhs0():
    """Test MB02RD with nrhs=0 (quick return)."""
    from slicot import mb02sd, mb02rd

    np.random.seed(456)
    n = 3
    nrhs = 0

    h = np.random.randn(n, n).astype(float, order='F')
    b = np.zeros((n, 1), order='F', dtype=float)
    ipiv = np.zeros(n, dtype=np.int32)

    x_out, info = mb02rd('N', n, nrhs, h, ipiv, b)

    assert info == 0


def test_mb02rd_invalid_trans():
    """Test MB02RD with invalid TRANS parameter."""
    from slicot import mb02rd

    n = 2
    nrhs = 1

    h = np.eye(n, order='F', dtype=float)
    b = np.ones((n, nrhs), order='F', dtype=float)
    ipiv = np.array([1, 2], dtype=np.int32)

    x_out, info = mb02rd('X', n, nrhs, h, ipiv, b)

    assert info == -1



"""
Tests for mb04rv - Complex generalized Sylvester equation solver

Solves:
    A * R - L * B = scale * C
    D * R - L * E = scale * F

where A, B, D, E are upper triangular complex matrices (generalized Schur form).
Solution (R, L) overwrites (C, F).
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose


def cabs1(z):
    """Taxi-cab norm: |real| + |imag| (as used in SLICOT)"""
    return np.abs(z.real) + np.abs(z.imag)


class TestMB04RV:
    """Tests for mb04rv complex generalized Sylvester solver."""

    def test_basic_1x1(self):
        """
        Test 1x1 system (simplest case).

        Random seed: 42 (for reproducibility)

        Eigenvalues of (A,D) and (B,E) must be distinct.
        (A,D) has eigenvalue A/D = (2+1j)/1 = 2+1j
        (B,E) has eigenvalue B/E = (3+0.5j)/1 = 3+0.5j
        These are distinct, so system is not singular.
        """
        from slicot import mb04rv

        np.random.seed(42)

        m, n = 1, 1
        pmax = 1e6

        a = np.array([[2.0 + 1.0j]], dtype=complex, order='F')
        b = np.array([[3.0 + 0.5j]], dtype=complex, order='F')
        d = np.array([[1.0 + 0.0j]], dtype=complex, order='F')
        e = np.array([[1.0 + 0.0j]], dtype=complex, order='F')

        c = np.array([[3.0 + 2.0j]], dtype=complex, order='F')
        f = np.array([[1.0 + 1.0j]], dtype=complex, order='F')
        c_orig = c.copy()
        f_orig = f.copy()

        r, l, scale, info = mb04rv(m, n, pmax, a, b, c, d, e, f)

        assert info == 0
        assert 0.0 < scale <= 1.0

        residual1 = a @ r - l @ b - scale * c_orig
        residual2 = d @ r - l @ e - scale * f_orig
        assert_allclose(residual1, 0.0, atol=1e-14)
        assert_allclose(residual2, 0.0, atol=1e-14)

    def test_2x2_system(self):
        """
        Test 2x2 upper triangular system.

        Random seed: 123 (for reproducibility)
        """
        from slicot import mb04rv

        np.random.seed(123)

        m, n = 2, 2
        pmax = 1e6

        a = np.array([[2.0+1.0j, 0.5+0.2j],
                      [0.0+0.0j, 1.0+0.5j]], dtype=complex, order='F')
        b = np.array([[1.0+0.3j, 0.3+0.1j],
                      [0.0+0.0j, 0.8+0.2j]], dtype=complex, order='F')
        d = np.array([[1.0+0.0j, 0.2+0.1j],
                      [0.0+0.0j, 0.5+0.0j]], dtype=complex, order='F')
        e = np.array([[0.5+0.0j, 0.1+0.0j],
                      [0.0+0.0j, 0.3+0.0j]], dtype=complex, order='F')

        c = np.array([[1.0+1.0j, 2.0+0.5j],
                      [0.5+0.2j, 1.5+1.0j]], dtype=complex, order='F')
        f = np.array([[0.5+0.5j, 1.0+0.3j],
                      [0.3+0.1j, 0.8+0.6j]], dtype=complex, order='F')
        c_orig = c.copy()
        f_orig = f.copy()

        r, l, scale, info = mb04rv(m, n, pmax, a, b, c, d, e, f)

        assert info == 0
        assert 0.0 < scale <= 1.0

        residual1 = a @ r - l @ b - scale * c_orig
        residual2 = d @ r - l @ e - scale * f_orig
        assert_allclose(residual1, 0.0, atol=1e-13)
        assert_allclose(residual2, 0.0, atol=1e-13)

    def test_rectangular_3x2(self):
        """
        Test 3x2 rectangular system.

        Random seed: 456 (for reproducibility)
        """
        from slicot import mb04rv

        np.random.seed(456)

        m, n = 3, 2
        pmax = 1e6

        a = np.array([[2.0+0.5j, 0.3+0.1j, 0.1+0.05j],
                      [0.0+0.0j, 1.5+0.3j, 0.2+0.1j],
                      [0.0+0.0j, 0.0+0.0j, 1.0+0.2j]], dtype=complex, order='F')
        b = np.array([[1.0+0.2j, 0.2+0.1j],
                      [0.0+0.0j, 0.8+0.1j]], dtype=complex, order='F')
        d = np.array([[1.0+0.0j, 0.1+0.05j, 0.05+0.02j],
                      [0.0+0.0j, 0.5+0.0j, 0.1+0.05j],
                      [0.0+0.0j, 0.0+0.0j, 0.3+0.0j]], dtype=complex, order='F')
        e = np.array([[0.5+0.0j, 0.1+0.0j],
                      [0.0+0.0j, 0.4+0.0j]], dtype=complex, order='F')

        c = np.random.randn(m, n) + 1j * np.random.randn(m, n)
        c = np.asfortranarray(c)
        f = np.random.randn(m, n) + 1j * np.random.randn(m, n)
        f = np.asfortranarray(f)
        c_orig = c.copy()
        f_orig = f.copy()

        r, l, scale, info = mb04rv(m, n, pmax, a, b, c, d, e, f)

        assert info == 0
        assert 0.0 < scale <= 1.0

        residual1 = a @ r - l @ b - scale * c_orig
        residual2 = d @ r - l @ e - scale * f_orig
        assert_allclose(residual1, 0.0, atol=1e-12)
        assert_allclose(residual2, 0.0, atol=1e-12)

    def test_rectangular_2x3(self):
        """
        Test 2x3 rectangular system.

        Random seed: 789 (for reproducibility)
        """
        from slicot import mb04rv

        np.random.seed(789)

        m, n = 2, 3
        pmax = 1e6

        a = np.array([[2.0+0.5j, 0.3+0.1j],
                      [0.0+0.0j, 1.5+0.3j]], dtype=complex, order='F')
        b = np.array([[1.0+0.2j, 0.2+0.1j, 0.1+0.05j],
                      [0.0+0.0j, 0.8+0.1j, 0.15+0.05j],
                      [0.0+0.0j, 0.0+0.0j, 0.6+0.1j]], dtype=complex, order='F')
        d = np.array([[1.0+0.0j, 0.1+0.05j],
                      [0.0+0.0j, 0.5+0.0j]], dtype=complex, order='F')
        e = np.array([[0.5+0.0j, 0.1+0.0j, 0.05+0.0j],
                      [0.0+0.0j, 0.4+0.0j, 0.08+0.0j],
                      [0.0+0.0j, 0.0+0.0j, 0.3+0.0j]], dtype=complex, order='F')

        c = np.random.randn(m, n) + 1j * np.random.randn(m, n)
        c = np.asfortranarray(c)
        f = np.random.randn(m, n) + 1j * np.random.randn(m, n)
        f = np.asfortranarray(f)
        c_orig = c.copy()
        f_orig = f.copy()

        r, l, scale, info = mb04rv(m, n, pmax, a, b, c, d, e, f)

        assert info == 0
        assert 0.0 < scale <= 1.0

        residual1 = a @ r - l @ b - scale * c_orig
        residual2 = d @ r - l @ e - scale * f_orig
        assert_allclose(residual1, 0.0, atol=1e-12)
        assert_allclose(residual2, 0.0, atol=1e-12)

    def test_pmax_threshold(self):
        """
        Test that INFO=1 is returned when solution exceeds PMAX.

        Use very small PMAX to trigger the threshold.
        Eigenvalues must be distinct (A/D != B/E) to avoid INFO=2.
        """
        from slicot import mb04rv

        np.random.seed(111)

        m, n = 1, 1
        pmax = 0.001  # Very small threshold

        a = np.array([[2.0 + 1.0j]], dtype=complex, order='F')
        b = np.array([[3.0 + 0.5j]], dtype=complex, order='F')
        d = np.array([[1.0 + 0.0j]], dtype=complex, order='F')
        e = np.array([[1.0 + 0.0j]], dtype=complex, order='F')

        c = np.array([[100.0 + 50.0j]], dtype=complex, order='F')
        f = np.array([[50.0 + 25.0j]], dtype=complex, order='F')

        r, l, scale, info = mb04rv(m, n, pmax, a, b, c, d, e, f)

        assert info == 1

    def test_zero_dimensions(self):
        """
        Test quick return for zero dimensions.
        """
        from slicot import mb04rv

        m, n = 0, 0
        pmax = 1e6

        a = np.array([], dtype=complex, order='F').reshape(0, 0)
        b = np.array([], dtype=complex, order='F').reshape(0, 0)
        c = np.array([], dtype=complex, order='F').reshape(0, 0)
        d = np.array([], dtype=complex, order='F').reshape(0, 0)
        e = np.array([], dtype=complex, order='F').reshape(0, 0)
        f = np.array([], dtype=complex, order='F').reshape(0, 0)

        r, l, scale, info = mb04rv(m, n, pmax, a, b, c, d, e, f)

        assert info == 0
        assert scale == 1.0

    def test_m_zero_n_nonzero(self):
        """
        Test M=0, N>0 case.
        """
        from slicot import mb04rv

        m, n = 0, 2
        pmax = 1e6

        a = np.array([], dtype=complex, order='F').reshape(0, 0)
        b = np.array([[1.0+0.1j, 0.2+0.05j],
                      [0.0+0.0j, 0.8+0.1j]], dtype=complex, order='F')
        c = np.array([], dtype=complex, order='F').reshape(0, 2)
        d = np.array([], dtype=complex, order='F').reshape(0, 0)
        e = np.array([[0.5+0.0j, 0.1+0.0j],
                      [0.0+0.0j, 0.3+0.0j]], dtype=complex, order='F')
        f = np.array([], dtype=complex, order='F').reshape(0, 2)

        r, l, scale, info = mb04rv(m, n, pmax, a, b, c, d, e, f)

        assert info == 0
        assert scale == 1.0

    def test_scale_factor_applied(self):
        """
        Verify that scale factor is properly applied to solution.

        When scale < 1, the returned R and L satisfy:
        A * R - L * B = scale * C_original
        D * R - L * E = scale * F_original

        Random seed: 222 (for reproducibility)
        """
        from slicot import mb04rv

        np.random.seed(222)

        m, n = 2, 2
        pmax = 1e6

        a = np.array([[2.0+1.0j, 0.5+0.2j],
                      [0.0+0.0j, 1.0+0.5j]], dtype=complex, order='F')
        b = np.array([[1.0+0.3j, 0.3+0.1j],
                      [0.0+0.0j, 0.8+0.2j]], dtype=complex, order='F')
        d = np.array([[1.0+0.0j, 0.2+0.1j],
                      [0.0+0.0j, 0.5+0.0j]], dtype=complex, order='F')
        e = np.array([[0.5+0.0j, 0.1+0.0j],
                      [0.0+0.0j, 0.3+0.0j]], dtype=complex, order='F')

        c = np.random.randn(m, n) + 1j * np.random.randn(m, n)
        c = np.asfortranarray(c)
        f = np.random.randn(m, n) + 1j * np.random.randn(m, n)
        f = np.asfortranarray(f)
        c_orig = c.copy()
        f_orig = f.copy()

        r, l, scale, info = mb04rv(m, n, pmax, a, b, c, d, e, f)

        assert info == 0

        lhs1 = a @ r - l @ b
        rhs1 = scale * c_orig
        lhs2 = d @ r - l @ e
        rhs2 = scale * f_orig

        assert_allclose(lhs1, rhs1, atol=1e-13)
        assert_allclose(lhs2, rhs2, atol=1e-13)

    def test_diagonal_only_matrices(self):
        """
        Test with purely diagonal matrices (simplest upper triangular).

        Random seed: 333 (for reproducibility)
        """
        from slicot import mb04rv

        np.random.seed(333)

        m, n = 3, 2
        pmax = 1e6

        a = np.diag([2.0+0.5j, 1.5+0.3j, 1.0+0.2j]).astype(complex, order='F')
        b = np.diag([1.0+0.2j, 0.8+0.1j]).astype(complex, order='F')
        d = np.diag([1.0, 0.5, 0.3]).astype(complex, order='F')
        e = np.diag([0.5, 0.4]).astype(complex, order='F')

        c = np.random.randn(m, n) + 1j * np.random.randn(m, n)
        c = np.asfortranarray(c)
        f = np.random.randn(m, n) + 1j * np.random.randn(m, n)
        f = np.asfortranarray(f)
        c_orig = c.copy()
        f_orig = f.copy()

        r, l, scale, info = mb04rv(m, n, pmax, a, b, c, d, e, f)

        assert info == 0
        assert 0.0 < scale <= 1.0

        residual1 = a @ r - l @ b - scale * c_orig
        residual2 = d @ r - l @ e - scale * f_orig
        assert_allclose(residual1, 0.0, atol=1e-13)
        assert_allclose(residual2, 0.0, atol=1e-13)

    def test_cabs1_property(self):
        """
        Verify understanding of cabs1 (taxi-cab norm).

        The SLICOT implementation uses |real| + |imag| for efficiency.
        """
        z1 = 3.0 + 4.0j
        z2 = -2.0 - 1.0j

        assert cabs1(z1) == 7.0  # |3| + |4|
        assert cabs1(z2) == 3.0  # |-2| + |-1|
        assert cabs1(np.array([z1, z2])).tolist() == [7.0, 3.0]



"""
Tests for SB08NY - Compute B(z) = A(1/z) * A(z) for discrete-time spectral factorization.

SB08NY computes the coefficients of the polynomial B(z) = A(1/z) * A(z)
where A(z) is a polynomial given in increasing powers of z.

Also computes an accuracy norm for the computed coefficients.
"""
import numpy as np
from slicot import sb08ny


def test_sb08ny_constant():
    """
    Test with constant polynomial A(z) = 3.

    A(1/z) = 3
    B(z) = A(1/z)*A(z) = 3*3 = 9
    B = [9]
    """
    a = np.array([3.0], dtype=float, order='F')

    b, epsb = sb08ny(a)

    assert b.shape == (1,)
    np.testing.assert_allclose(b[0], 9.0, rtol=1e-14)
    assert epsb > 0


def test_sb08ny_linear():
    """
    Test with linear polynomial A(z) = 1 + 2*z.

    A(1/z) = 1 + 2/z = (z + 2)/z
    B(z) = A(1/z)*A(z) = (1 + 2/z)(1 + 2z) = 1 + 2z + 2/z + 4
         = 5 + 2z + 2/z = (2 + 5z + 2z^2)/z

    Coefficients of B(z) in increasing powers of z (DA+1 = 2):
    B[i] = sum_{k=0}^{DA-i+1} A[k] * A[i+k-1] for Fortran 1-based
    For C 0-based with DA=1:
    B[0] = A[0]*A[0] + A[1]*A[1] = 1*1 + 2*2 = 5
    B[1] = A[0]*A[1] = 1*2 = 2
    """
    a = np.array([1.0, 2.0], dtype=float, order='F')

    b, epsb = sb08ny(a)

    assert b.shape == (2,)
    np.testing.assert_allclose(b[0], 5.0, rtol=1e-14)
    np.testing.assert_allclose(b[1], 2.0, rtol=1e-14)


def test_sb08ny_quadratic():
    """
    Test with quadratic polynomial A(z) = 1 + 2*z + 3*z^2.

    For DA=2, B has 3 coefficients.
    Using the dot product formula B(I) = DDOT(DA-I+2, A(1), 1, A(I), 1):
    B[0] = A[0]*A[0] + A[1]*A[1] + A[2]*A[2] = 1 + 4 + 9 = 14
    B[1] = A[0]*A[1] + A[1]*A[2] = 1*2 + 2*3 = 8
    B[2] = A[0]*A[2] = 1*3 = 3
    """
    a = np.array([1.0, 2.0, 3.0], dtype=float, order='F')

    b, epsb = sb08ny(a)

    assert b.shape == (3,)
    np.testing.assert_allclose(b[0], 14.0, rtol=1e-14)
    np.testing.assert_allclose(b[1], 8.0, rtol=1e-14)
    np.testing.assert_allclose(b[2], 3.0, rtol=1e-14)


def test_sb08ny_property_autocorrelation():
    """
    Verify B(z) = A(1/z) * A(z) represents autocorrelation coefficients.

    For polynomial A(z) = sum_{k=0}^{n} a_k * z^k,
    B(z) = sum_{k=-n}^{n} r_k * z^k where r_k = sum_{j} a_j * a_{j+|k|}

    The output B contains r_0, r_1, ..., r_n (non-negative lags).
    Due to symmetry r_k = r_{-k}.

    Property: B(0) = ||A||^2 = sum of squares of coefficients.

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    for _ in range(10):
        da = np.random.randint(1, 8)
        a = np.random.randn(da + 1)
        a = np.array(a, dtype=float, order='F')

        b, epsb = sb08ny(a.copy(order='F'))

        # B[0] = sum of squared coefficients (autocorrelation at lag 0)
        expected_b0 = np.sum(a ** 2)
        np.testing.assert_allclose(b[0], expected_b0, rtol=1e-14)


def test_sb08ny_property_polynomial_evaluation():
    """
    Verify B(z) = A(1/z) * A(z) by evaluating at specific points.

    For z != 0, evaluate A(1/z)*A(z) and compare with polynomial B.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    da = 3
    a = np.random.randn(da + 1)
    a = np.array(a, dtype=float, order='F')

    b, epsb = sb08ny(a.copy(order='F'))

    # Test points (avoiding z=0)
    for z in [0.5, 1.0, 1.5, 2.0, -0.5, -1.0]:
        # A(z) = sum_{k=0}^{da} a[k] * z^k
        a_z = sum(a[k] * (z ** k) for k in range(da + 1))
        # A(1/z) = sum_{k=0}^{da} a[k] * (1/z)^k
        a_inv_z = sum(a[k] * (z ** (-k)) for k in range(da + 1))
        # Product
        product = a_z * a_inv_z

        # B(z) = sum_{k=-da}^{da} r_k * z^k
        # With r_k = r_{-k}, we have:
        # B(z) = r_0 + sum_{k=1}^{da} r_k * (z^k + z^{-k})
        b_z = b[0]
        for k in range(1, da + 1):
            b_z += b[k] * (z ** k + z ** (-k))

        np.testing.assert_allclose(b_z, product, rtol=1e-12)


def test_sb08ny_symmetry():
    """
    Verify autocorrelation symmetry property.

    B(z) has the property that coefficient of z^k equals coefficient of z^{-k}.
    The output array contains only non-negative lag coefficients.

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    da = 4
    a = np.random.randn(da + 1)
    a = np.array(a, dtype=float, order='F')

    b, epsb = sb08ny(a.copy(order='F'))

    # Verify by explicit autocorrelation computation
    # r_k = sum_{j=0}^{n-k} a[j] * a[j+k]
    for k in range(da + 1):
        expected_rk = sum(a[j] * a[j + k] for j in range(da + 1 - k))
        np.testing.assert_allclose(b[k], expected_rk, rtol=1e-14)


def test_sb08ny_zero_polynomial():
    """
    Test edge case with zero polynomial A(z) = 0.

    B(z) = 0 * 0 = 0
    """
    a = np.array([0.0], dtype=float, order='F')

    b, epsb = sb08ny(a)

    assert b.shape == (1,)
    np.testing.assert_allclose(b[0], 0.0, atol=1e-14)


def test_sb08ny_accuracy_norm():
    """
    Test that accuracy norm is computed correctly.

    EPSB = 3 * machine_eps * B[0]
    where B[0] = sum of squares of coefficients.

    Note: DLAMCH('E') returns machine epsilon = eps/2 (unit roundoff),
    not numpy's eps. This is the relative machine precision b^(1-p).
    """
    a = np.array([1.0, 2.0, 3.0], dtype=float, order='F')

    b, epsb = sb08ny(a)

    # DLAMCH('E') returns machine epsilon = eps/2 (unit roundoff)
    machine_eps = np.finfo(float).eps / 2.0
    expected_epsb = 3.0 * machine_eps * b[0]
    np.testing.assert_allclose(epsb, expected_epsb, rtol=1e-14)


def test_sb08ny_numerical_example():
    """
    Test with specific numerical example.

    A(z) = 2 + 3*z
    A(1/z) = 2 + 3/z
    B(z) = (2 + 3/z)(2 + 3z) = 4 + 6z + 6/z + 9 = 13 + 6z + 6/z

    Autocorrelation: r_0 = 2^2 + 3^2 = 13, r_1 = 2*3 = 6
    """
    a = np.array([2.0, 3.0], dtype=float, order='F')

    b, epsb = sb08ny(a)

    np.testing.assert_allclose(b[0], 13.0, rtol=1e-14)
    np.testing.assert_allclose(b[1], 6.0, rtol=1e-14)



"""Tests for MB04HD - Reducing skew-Hamiltonian/Hamiltonian pencil to generalized Schur form."""

import numpy as np
import pytest


def test_mb04hd_basic():
    """
    Test MB04HD with COMPQ1='I', COMPQ2='I'.

    Tests transformation of block diagonal/anti-diagonal pencil to generalized Schur form.
    The input pencil has structure:
      A = diag(A11, A22) with A11, A22 upper triangular
      B = anti-diag(B21, B12) with B12 upper triangular, B21 upper quasi-triangular

    N must be even (N >= 0).

    Random seed: 42 (for reproducibility)
    """
    from slicot import mb04hd

    np.random.seed(42)
    n = 4
    m = n // 2

    a11 = np.triu(np.random.randn(m, m))
    a22 = np.triu(np.random.randn(m, m))
    a = np.zeros((n, n), order='F', dtype=float)
    a[:m, :m] = a11
    a[m:, m:] = a22

    b12 = np.triu(np.random.randn(m, m))
    b21 = np.triu(np.random.randn(m, m))
    b = np.zeros((n, n), order='F', dtype=float)
    b[:m, m:] = b12
    b[m:, :m] = b21

    result = mb04hd('I', 'I', a, b)
    a_out, b_out, q1, q2, info = result

    assert info == 0 or info in [1, 2, 3, 4]

    assert a_out.shape == (n, n)
    assert b_out.shape == (n, n)
    assert q1.shape == (n, n)
    assert q2.shape == (n, n)


def test_mb04hd_orthogonality():
    """
    Validate Q1 and Q2 are orthogonal matrices.

    For COMPQ1='I', COMPQ2='I', the transformation matrices
    Q1 and Q2 must satisfy Q'Q = QQ' = I.

    Random seed: 123 (for reproducibility)
    """
    from slicot import mb04hd

    np.random.seed(123)
    n = 6
    m = n // 2

    a11 = np.triu(np.random.randn(m, m))
    a22 = np.triu(np.random.randn(m, m))
    a = np.zeros((n, n), order='F', dtype=float)
    a[:m, :m] = a11
    a[m:, m:] = a22

    b12 = np.triu(np.random.randn(m, m))
    b21 = np.triu(np.random.randn(m, m))
    b = np.zeros((n, n), order='F', dtype=float)
    b[:m, m:] = b12
    b[m:, :m] = b21

    result = mb04hd('I', 'I', a, b)
    a_out, b_out, q1, q2, info = result

    assert info == 0 or info in [1, 2, 3, 4]

    if info == 0:
        np.testing.assert_allclose(q1 @ q1.T, np.eye(n), rtol=1e-12, atol=1e-12)
        np.testing.assert_allclose(q2 @ q2.T, np.eye(n), rtol=1e-12, atol=1e-12)
        np.testing.assert_allclose(q1.T @ q1, np.eye(n), rtol=1e-12, atol=1e-12)
        np.testing.assert_allclose(q2.T @ q2, np.eye(n), rtol=1e-12, atol=1e-12)


def test_mb04hd_upper_triangular_a():
    """
    Validate transformed A is upper triangular.

    After transformation, Q2' A Q1 should be upper triangular.

    Random seed: 456 (for reproducibility)
    """
    from slicot import mb04hd

    np.random.seed(456)
    n = 4
    m = n // 2

    a11 = np.triu(np.random.randn(m, m))
    a22 = np.triu(np.random.randn(m, m))
    a = np.zeros((n, n), order='F', dtype=float)
    a[:m, :m] = a11
    a[m:, m:] = a22

    b12 = np.triu(np.random.randn(m, m))
    b21 = np.triu(np.random.randn(m, m))
    b = np.zeros((n, n), order='F', dtype=float)
    b[:m, m:] = b12
    b[m:, :m] = b21

    result = mb04hd('I', 'I', a, b)
    a_out, b_out, q1, q2, info = result

    assert info == 0 or info in [1, 2, 3, 4]

    if info == 0:
        np.testing.assert_allclose(np.tril(a_out, -1), 0, atol=1e-10)


def test_mb04hd_n_zero():
    """Test MB04HD with N=0 (quick return)."""
    from slicot import mb04hd

    a = np.array([], dtype=float, order='F').reshape(0, 0)
    b = np.array([], dtype=float, order='F').reshape(0, 0)

    result = mb04hd('N', 'N', a, b)
    a_out, b_out, q1, q2, info = result

    assert info == 0


def test_mb04hd_no_orthogonal_matrices():
    """
    Test MB04HD with COMPQ1='N', COMPQ2='N'.

    Q1 and Q2 are not computed.

    Random seed: 789 (for reproducibility)
    """
    from slicot import mb04hd

    np.random.seed(789)
    n = 4
    m = n // 2

    a11 = np.triu(np.random.randn(m, m))
    a22 = np.triu(np.random.randn(m, m))
    a = np.zeros((n, n), order='F', dtype=float)
    a[:m, :m] = a11
    a[m:, m:] = a22

    b12 = np.triu(np.random.randn(m, m))
    b21 = np.triu(np.random.randn(m, m))
    b = np.zeros((n, n), order='F', dtype=float)
    b[:m, m:] = b12
    b[m:, :m] = b21

    result = mb04hd('N', 'N', a, b)
    a_out, b_out, q1, q2, info = result

    assert info == 0 or info in [1, 2, 3, 4]
    assert a_out.shape == (n, n)
    assert b_out.shape == (n, n)


def test_mb04hd_small_n2():
    """
    Test MB04HD with N=2 (smallest even case).

    Random seed: 111 (for reproducibility)
    """
    from slicot import mb04hd

    np.random.seed(111)
    n = 2
    m = n // 2

    a = np.zeros((n, n), order='F', dtype=float)
    a[0, 0] = 1.5
    a[1, 1] = 2.5

    b = np.zeros((n, n), order='F', dtype=float)
    b[0, 1] = 0.3
    b[1, 0] = 0.4

    result = mb04hd('I', 'I', a, b)
    a_out, b_out, q1, q2, info = result

    assert info == 0 or info in [1, 2, 3, 4]
    assert a_out.shape == (n, n)
    assert b_out.shape == (n, n)


def test_mb04hd_update_mode():
    """
    Test MB04HD with COMPQ1='U', COMPQ2='U' (update mode).

    When update mode, initial orthogonal matrices are provided and multiplied
    by the transformation matrices.

    Random seed: 222 (for reproducibility)
    """
    from slicot import mb04hd

    np.random.seed(222)
    n = 4
    m = n // 2

    a11 = np.triu(np.random.randn(m, m))
    a22 = np.triu(np.random.randn(m, m))
    a = np.zeros((n, n), order='F', dtype=float)
    a[:m, :m] = a11
    a[m:, m:] = a22

    b12 = np.triu(np.random.randn(m, m))
    b21 = np.triu(np.random.randn(m, m))
    b = np.zeros((n, n), order='F', dtype=float)
    b[:m, m:] = b12
    b[m:, :m] = b21

    q1_init = np.eye(n, dtype=float, order='F')
    q2_init = np.eye(n, dtype=float, order='F')

    result = mb04hd('U', 'U', a, b, q1=q1_init.copy(), q2=q2_init.copy())
    a_out, b_out, q1, q2, info = result

    assert info == 0 or info in [1, 2, 3, 4]

    if info == 0:
        np.testing.assert_allclose(q1 @ q1.T, np.eye(n), rtol=1e-12, atol=1e-12)
        np.testing.assert_allclose(q2 @ q2.T, np.eye(n), rtol=1e-12, atol=1e-12)


def test_mb04hd_transformation_property():
    """
    Validate transformation: Q2' * A_in * Q1 = A_out and Q2' * B_in * Q1 = B_out.

    Random seed: 333 (for reproducibility)
    """
    from slicot import mb04hd

    np.random.seed(333)
    n = 4
    m = n // 2

    a11 = np.triu(np.random.randn(m, m))
    a22 = np.triu(np.random.randn(m, m))
    a_in = np.zeros((n, n), order='F', dtype=float)
    a_in[:m, :m] = a11
    a_in[m:, m:] = a22
    a_copy = a_in.copy()

    b12 = np.triu(np.random.randn(m, m))
    b21 = np.triu(np.random.randn(m, m))
    b_in = np.zeros((n, n), order='F', dtype=float)
    b_in[:m, m:] = b12
    b_in[m:, :m] = b21
    b_copy = b_in.copy()

    result = mb04hd('I', 'I', a_in, b_in)
    a_out, b_out, q1, q2, info = result

    assert info == 0 or info in [1, 2, 3, 4]

    if info == 0:
        a_transformed = q2.T @ a_copy @ q1
        np.testing.assert_allclose(a_transformed, a_out, rtol=1e-10, atol=1e-10)

        b_transformed = q2.T @ b_copy @ q1
        np.testing.assert_allclose(b_transformed, b_out, rtol=1e-10, atol=1e-10)


def test_mb04hd_larger_n():
    """
    Test MB04HD with larger N=8.

    Random seed: 444 (for reproducibility)
    """
    from slicot import mb04hd

    np.random.seed(444)
    n = 8
    m = n // 2

    a11 = np.triu(np.random.randn(m, m))
    a22 = np.triu(np.random.randn(m, m))
    a = np.zeros((n, n), order='F', dtype=float)
    a[:m, :m] = a11
    a[m:, m:] = a22

    b12 = np.triu(np.random.randn(m, m))
    b21 = np.triu(np.random.randn(m, m))
    b = np.zeros((n, n), order='F', dtype=float)
    b[:m, m:] = b12
    b[m:, :m] = b21

    result = mb04hd('I', 'I', a, b)
    a_out, b_out, q1, q2, info = result

    assert info == 0 or info in [1, 2, 3, 4]

    if info == 0:
        np.testing.assert_allclose(q1 @ q1.T, np.eye(n), rtol=1e-11, atol=1e-11)
        np.testing.assert_allclose(q2 @ q2.T, np.eye(n), rtol=1e-11, atol=1e-11)
        np.testing.assert_allclose(np.tril(a_out, -1), 0, atol=1e-10)



"""
Tests for MB04WR: Generate orthogonal symplectic matrices U or V from symplectic reflectors.

MB04WR generates orthogonal symplectic matrices U or V, defined as products of
symplectic reflectors and Givens rotations, as returned by MB04TS or MB04TB.

The matrices U and V are returned in terms of their first N/2 rows:
    U = [[U1, U2], [-U2, U1]]
    V = [[V1, V2], [-V2, V1]]
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose

from slicot import mb04ts, mb04wr


class TestMB04WRBasic:
    """Basic functionality tests for mb04wr."""

    def test_job_u_trans_n_basic(self):
        """
        Test JOB='U', TRANS='N' case with data from MB04TS.

        For JOB='U', TRANS='N':
        - Q1 input: A_out contains FU reflector vectors (in columns)
        - Q2 input: Q_out contains HU reflector vectors (in columns) with tau on diagonal

        Random seed: 42 (for reproducibility)
        """
        np.random.seed(42)
        n = 4
        ilo = 1

        A = np.random.randn(n, n).astype(float, order='F')
        B = np.random.randn(n, n).astype(float, order='F')
        G = np.random.randn(n, n)
        G = (G + G.T).astype(float, order='F')
        Q = np.random.randn(n, n).astype(float, order='F')

        A_out, B_out, G_out, Q_out, csl, csr, taul, taur, info = mb04ts(
            'N', 'N', n, ilo, A.copy(), B.copy(), G.copy(), Q.copy()
        )
        assert info == 0

        q1 = A_out.copy()
        q2 = Q_out.copy()

        q1_out, q2_out, info = mb04wr('U', 'N', n, ilo, q1, q2, csl, taul)

        assert info == 0, f"Expected info=0, got {info}"
        assert q1_out.shape == (n, n), f"Expected Q1 shape {(n, n)}, got {q1_out.shape}"
        assert q2_out.shape == (n, n), f"Expected Q2 shape {(n, n)}, got {q2_out.shape}"

    def test_job_u_trans_t_basic(self):
        """
        Test JOB='U', TRANS='T' case.

        For JOB='U', TRANS='T':
        - Q1 input: A_out contains FU reflector vectors (in rows when TRANA='T')
        - Q2 input: Q_out contains HU reflector vectors (in columns) with tau on diagonal

        Random seed: 123 (for reproducibility)
        """
        np.random.seed(123)
        n = 4
        ilo = 1

        A = np.random.randn(n, n).astype(float, order='F')
        B = np.random.randn(n, n).astype(float, order='F')
        G = np.random.randn(n, n)
        G = (G + G.T).astype(float, order='F')
        Q = np.random.randn(n, n).astype(float, order='F')

        A_out, B_out, G_out, Q_out, csl, csr, taul, taur, info = mb04ts(
            'T', 'N', n, ilo, A.copy(), B.copy(), G.copy(), Q.copy()
        )
        assert info == 0

        q1 = A_out.copy()
        q2 = Q_out.copy()

        q1_out, q2_out, info = mb04wr('U', 'T', n, ilo, q1, q2, csl, taul)

        assert info == 0, f"Expected info=0, got {info}"
        assert q1_out.shape == (n, n)
        assert q2_out.shape == (n, n)

    def test_job_v_trans_n_basic(self):
        """
        Test JOB='V', TRANS='N' case.

        For JOB='V', TRANS='N':
        - Q1 input: B_out contains FV reflector vectors (in rows when TRANB='N')
        - Q2 input: Q_out contains HV reflector vectors (in rows)

        Random seed: 456 (for reproducibility)
        """
        np.random.seed(456)
        n = 4
        ilo = 1

        A = np.random.randn(n, n).astype(float, order='F')
        B = np.random.randn(n, n).astype(float, order='F')
        G = np.random.randn(n, n)
        G = (G + G.T).astype(float, order='F')
        Q = np.random.randn(n, n).astype(float, order='F')

        A_out, B_out, G_out, Q_out, csl, csr, taul, taur, info = mb04ts(
            'N', 'N', n, ilo, A.copy(), B.copy(), G.copy(), Q.copy()
        )
        assert info == 0

        q1 = B_out.copy()
        q2 = Q_out.copy()

        q1_out, q2_out, info = mb04wr('V', 'N', n, ilo, q1, q2, csr, taur)

        assert info == 0, f"Expected info=0, got {info}"
        assert q1_out.shape == (n, n)
        assert q2_out.shape == (n, n)

    def test_job_v_trans_t_basic(self):
        """
        Test JOB='V', TRANS='T' case.

        For JOB='V', TRANS='T':
        - Q1 input: B_out contains FV reflector vectors (in columns when TRANB='T')
        - Q2 input: Q_out contains HV reflector vectors (in rows)

        Random seed: 789 (for reproducibility)
        """
        np.random.seed(789)
        n = 4
        ilo = 1

        A = np.random.randn(n, n).astype(float, order='F')
        B = np.random.randn(n, n).astype(float, order='F')
        G = np.random.randn(n, n)
        G = (G + G.T).astype(float, order='F')
        Q = np.random.randn(n, n).astype(float, order='F')

        A_out, B_out, G_out, Q_out, csl, csr, taul, taur, info = mb04ts(
            'N', 'T', n, ilo, A.copy(), B.copy(), G.copy(), Q.copy()
        )
        assert info == 0

        q1 = B_out.copy()
        q2 = Q_out.copy()

        q1_out, q2_out, info = mb04wr('V', 'T', n, ilo, q1, q2, csr, taur)

        assert info == 0, f"Expected info=0, got {info}"
        assert q1_out.shape == (n, n)
        assert q2_out.shape == (n, n)


class TestMB04WREdgeCases:
    """Edge case tests for mb04wr."""

    def test_n_equals_0(self):
        """Test with N=0 (quick return)."""
        n = 0
        ilo = 1
        q1 = np.zeros((0, 0), dtype=float, order='F')
        q2 = np.zeros((0, 0), dtype=float, order='F')
        cs = np.zeros(0, dtype=float)
        tau = np.zeros(0, dtype=float)

        q1_out, q2_out, info = mb04wr('U', 'N', n, ilo, q1, q2, cs, tau)

        assert info == 0

    def test_n_equals_1(self):
        """Test with N=1 (minimal case)."""
        np.random.seed(111)
        n = 1
        ilo = 1

        A = np.random.randn(n, n).astype(float, order='F')
        B = np.random.randn(n, n).astype(float, order='F')
        G = np.random.randn(n, n)
        G = (G + G.T).astype(float, order='F')
        Q = np.random.randn(n, n).astype(float, order='F')

        A_out, B_out, G_out, Q_out, csl, csr, taul, taur, info = mb04ts(
            'N', 'N', n, ilo, A.copy(), B.copy(), G.copy(), Q.copy()
        )
        assert info == 0

        q1 = A_out.copy()
        q2 = Q_out.copy()

        q1_out, q2_out, info = mb04wr('U', 'N', n, ilo, q1, q2, csl, taul)

        assert info == 0


class TestMB04WRMathematicalProperties:
    """Mathematical property tests for mb04wr."""

    def test_orthogonality_job_u(self):
        """
        Verify orthogonality property U1^T U1 + U2^T U2 = I for JOB='U'.

        Random seed: 1234 (for reproducibility)
        """
        np.random.seed(1234)
        n = 4
        ilo = 1

        A = np.random.randn(n, n).astype(float, order='F')
        B = np.random.randn(n, n).astype(float, order='F')
        G = np.random.randn(n, n)
        G = (G + G.T).astype(float, order='F')
        Q = np.random.randn(n, n).astype(float, order='F')

        A_out, B_out, G_out, Q_out, csl, csr, taul, taur, info = mb04ts(
            'N', 'N', n, ilo, A.copy(), B.copy(), G.copy(), Q.copy()
        )
        assert info == 0

        q1 = A_out.copy()
        q2 = Q_out.copy()

        u1, u2, info = mb04wr('U', 'N', n, ilo, q1, q2, csl, taul)
        assert info == 0

        result = u1.T @ u1 + u2.T @ u2
        assert_allclose(result, np.eye(n), rtol=1e-13, atol=1e-14,
                       err_msg="U1^T U1 + U2^T U2 should equal I")

    def test_symplectic_structure_job_u(self):
        """
        Verify full U preserves symplectic structure: U^T J U = J for JOB='U'.

        J = [[0, I], [-I, 0]] is the standard symplectic form.
        Random seed: 5678 (for reproducibility)
        """
        np.random.seed(5678)
        n = 3
        ilo = 1

        A = np.random.randn(n, n).astype(float, order='F')
        B = np.random.randn(n, n).astype(float, order='F')
        G = np.random.randn(n, n)
        G = (G + G.T).astype(float, order='F')
        Q = np.random.randn(n, n).astype(float, order='F')

        A_out, B_out, G_out, Q_out, csl, csr, taul, taur, info = mb04ts(
            'N', 'N', n, ilo, A.copy(), B.copy(), G.copy(), Q.copy()
        )
        assert info == 0

        q1 = A_out.copy()
        q2 = Q_out.copy()

        u1, u2, info = mb04wr('U', 'N', n, ilo, q1, q2, csl, taul)
        assert info == 0

        u_full = np.block([
            [u1, u2],
            [-u2, u1]
        ])

        j_mat = np.block([
            [np.zeros((n, n)), np.eye(n)],
            [-np.eye(n), np.zeros((n, n))]
        ])

        result = u_full.T @ j_mat @ u_full
        assert_allclose(result, j_mat, rtol=1e-13, atol=1e-14,
                       err_msg="U should preserve symplectic structure")

    def test_isotropic_columns(self):
        """
        Verify that columns satisfy isotropic condition: U1^T U2 = U2^T U1.

        Random seed: 9012 (for reproducibility)
        """
        np.random.seed(9012)
        n = 4
        ilo = 1

        A = np.random.randn(n, n).astype(float, order='F')
        B = np.random.randn(n, n).astype(float, order='F')
        G = np.random.randn(n, n)
        G = (G + G.T).astype(float, order='F')
        Q = np.random.randn(n, n).astype(float, order='F')

        A_out, B_out, G_out, Q_out, csl, csr, taul, taur, info = mb04ts(
            'N', 'N', n, ilo, A.copy(), B.copy(), G.copy(), Q.copy()
        )
        assert info == 0

        q1 = A_out.copy()
        q2 = Q_out.copy()

        u1, u2, info = mb04wr('U', 'N', n, ilo, q1, q2, csl, taul)
        assert info == 0

        u1_t_u2 = u1.T @ u2
        u2_t_u1 = u2.T @ u1
        assert_allclose(u1_t_u2, u2_t_u1, rtol=1e-13, atol=1e-14,
                       err_msg="Columns should be isotropic: U1^T U2 = U2^T U1")

    def test_full_orthogonality(self):
        """
        Verify full U matrix is orthogonal: U^T U = I.

        Random seed: 3456 (for reproducibility)
        """
        np.random.seed(3456)
        n = 4
        ilo = 1

        A = np.random.randn(n, n).astype(float, order='F')
        B = np.random.randn(n, n).astype(float, order='F')
        G = np.random.randn(n, n)
        G = (G + G.T).astype(float, order='F')
        Q = np.random.randn(n, n).astype(float, order='F')

        A_out, B_out, G_out, Q_out, csl, csr, taul, taur, info = mb04ts(
            'N', 'N', n, ilo, A.copy(), B.copy(), G.copy(), Q.copy()
        )
        assert info == 0

        q1 = A_out.copy()
        q2 = Q_out.copy()

        u1, u2, info = mb04wr('U', 'N', n, ilo, q1, q2, csl, taul)
        assert info == 0

        u_full = np.block([
            [u1, u2],
            [-u2, u1]
        ])

        utu = u_full.T @ u_full
        assert_allclose(utu, np.eye(2 * n), rtol=1e-13, atol=1e-14,
                       err_msg="Full U should be orthogonal")


class TestMB04WRErrors:
    """Error handling tests for mb04wr."""

    def test_invalid_job(self):
        """Test that invalid JOB returns error."""
        n = 3
        ilo = 1
        q1 = np.zeros((n, n), dtype=float, order='F')
        q2 = np.zeros((n, n), dtype=float, order='F')
        cs = np.zeros(2 * n, dtype=float)
        tau = np.zeros(n, dtype=float)

        with pytest.raises(ValueError):
            mb04wr('X', 'N', n, ilo, q1, q2, cs, tau)

    def test_invalid_trans(self):
        """Test that invalid TRANS returns error."""
        n = 3
        ilo = 1
        q1 = np.zeros((n, n), dtype=float, order='F')
        q2 = np.zeros((n, n), dtype=float, order='F')
        cs = np.zeros(2 * n, dtype=float)
        tau = np.zeros(n, dtype=float)

        with pytest.raises(ValueError):
            mb04wr('U', 'X', n, ilo, q1, q2, cs, tau)

    def test_negative_n(self):
        """Test that negative N returns error."""
        n = -1
        ilo = 1
        q1 = np.zeros((1, 1), dtype=float, order='F')
        q2 = np.zeros((1, 1), dtype=float, order='F')
        cs = np.zeros(2, dtype=float)
        tau = np.zeros(1, dtype=float)

        with pytest.raises(ValueError):
            mb04wr('U', 'N', n, ilo, q1, q2, cs, tau)

    def test_ilo_out_of_range_low(self):
        """Test that ILO < 1 returns error."""
        n = 3
        ilo = 0
        q1 = np.zeros((n, n), dtype=float, order='F')
        q2 = np.zeros((n, n), dtype=float, order='F')
        cs = np.zeros(2 * n, dtype=float)
        tau = np.zeros(n, dtype=float)

        with pytest.raises(ValueError):
            mb04wr('U', 'N', n, ilo, q1, q2, cs, tau)

    def test_ilo_out_of_range_high(self):
        """Test that ILO > N returns error."""
        n = 3
        ilo = 5
        q1 = np.zeros((n, n), dtype=float, order='F')
        q2 = np.zeros((n, n), dtype=float, order='F')
        cs = np.zeros(2 * n, dtype=float)
        tau = np.zeros(n, dtype=float)

        with pytest.raises(ValueError):
            mb04wr('U', 'N', n, ilo, q1, q2, cs, tau)
